package org.opensourcephysics.tuleja.doublependulum3d;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import org.opensourcephysics.display3d.*;
import org.opensourcephysics.display3d.*;
import org.opensourcephysics.numerics.*;

public class JDoublePendulum3d
    extends JApplet
    implements ActionListener, ChangeListener {

  Container cp;
  JPanel pnlSouth, pnlSliders, pnlButton;
  JButton btnStartStop;
  JSlider slPhi1, slPhi2, slw1, slw2;
  DrawingPanel3D panel3D;
  ActionListener tickingHandler;

  DShapeGroup upperPart, stationaryPart, lowerPart;
  DBox pedestala, pedestalb, bar1a, bar1b, bar2, base;
  Dataset3D dataset;

  DCylinder axlea, axleb, axlec;
  DoublePendulumODE ode = new DoublePendulum();
  ODEAdaptiveSolver odeAdaptiveSolver = new RK45MultiStep(ode);
  javax.swing.Timer timer = new javax.swing.Timer(100, this);
  double L1display, L2display, tbase, wbase, hpedestal, wpedestal, d, gap, r;
  double phi1, phi2, w1, w2, t, dt, L1, L2;
  boolean trace;


  public static void main(String[] args) {
      JFrame fr = new JFrame("JDoublePendulum");
      fr.setSize(400, 550);
      JDoublePendulum3d applet = new JDoublePendulum3d();
      applet.init();
      applet.start();
      fr.getContentPane().add(applet, BorderLayout.CENTER);
      fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      fr.setVisible(true);
    }

  /**
   *
   */
  public void init() {
    //Creating GUI
    panel3D = new DrawingPanel3D();
    panel3D.setBorder(BorderFactory.createTitledBorder("3D Universe"));

    //button
    btnStartStop = new JButton("Start");
    btnStartStop.addActionListener(this);

    //sliders
    slPhi1 = new JSlider(JSlider.HORIZONTAL, 0, 360, 270);
    slPhi1.setBorder(BorderFactory.createTitledBorder("\u03D5_1 = " +
        slPhi1.getValue() + " "));
    slPhi1.addChangeListener(this);

    slPhi2 = new JSlider(JSlider.HORIZONTAL, 0, 360, 270);
    slPhi2.setBorder(BorderFactory.createTitledBorder("\u03D5_2 = " +
        slPhi2.getValue() + " "));
    slPhi2.addChangeListener(this);

    slw1 = new JSlider(JSlider.HORIZONTAL, -720, 720, 0);
    slw1.setBorder(BorderFactory.createTitledBorder("d\u03D5_1 / dt = " +
        slw1.getValue() + " /s"));
    slw1.addChangeListener(this);

    slw2 = new JSlider(JSlider.HORIZONTAL, -720, 720, 0);
    slw2.setBorder(BorderFactory.createTitledBorder("d\u03D5_2 / dt = " +
        slw2.getValue() + " /s"));
    slw2.addChangeListener(this);

    pnlSliders = new JPanel();
    pnlSliders.setBorder(BorderFactory.createTitledBorder("Initial conditions:"));
    pnlSliders.setLayout(new GridLayout(2, 2));
    pnlSliders.add(slPhi1);
    pnlSliders.add(slPhi2);
    pnlSliders.add(slw1);
    pnlSliders.add(slw2);

    pnlButton = new JPanel();
    pnlButton.setLayout(new GridLayout(1,1));
    pnlButton.add(btnStartStop);

    pnlSouth = new JPanel();
    pnlSouth.setLayout(new BoxLayout(pnlSouth, BoxLayout.Y_AXIS));
    pnlSouth.add(pnlSliders);
    pnlSouth.add(pnlButton);

    cp = getContentPane();

    cp.setLayout(new BorderLayout());
    cp.add(panel3D, BorderLayout.CENTER);
    cp.add(pnlSouth, BorderLayout.SOUTH);
    cp.validate();
    cp.repaint();
    initializeAnimation();

    tickingHandler = new TickingHandler(this);
    timer = new javax.swing.Timer(100, tickingHandler);
  }

  /**
   *
   */
  public void start() {}

  /**
   * This is done each time the Timer ticks
   * @param evt ActionEvent
   */
  public void actionPerformed(ActionEvent evt) {
    if (evt.getActionCommand().equals("Start")) {
      btnStartStop.setText("Stop");
      slPhi1.setEnabled(false);
      slPhi2.setEnabled(false);
      slw1.setEnabled(false);
      slw2.setEnabled(false);
      startAnimation();
    }
    else if (evt.getActionCommand().equals("Stop")) {
      btnStartStop.setText("Start");
      stopAnimation();
      slPhi1.setEnabled(true);
      slPhi2.setEnabled(true);
      slw1.setEnabled(true);
      slw2.setEnabled(true);
    }
  }

  /**
   *
   */
  public void create3DScene() {
    panel3D.clear();

    //Object of the virtual universe are set up here...
    d = 0.05; // thickness of each bar
    gap = 0.05; // distance between two box parts parts connected by an axel
    r = 0.01; // radius of axels
    L1display = L1 + 3 * r; // show upper assembly a bit longer than physical, to overlap axle
    L2display = L2 + 3 * r; // show lower bar a bit longer than physical, to overlap axle
    hpedestal = 1.3 * (L1 + L2); // height of pedestal
    wpedestal = 0.1; // width of pedestal
    tbase = 0.05; // thickness of base
    wbase = 1; // width of base

    Color pedestalColor = new Color(0.4f, 0.4f, 0.5f);

    base = new DBox(0, 0, 0, wbase, tbase, wbase, pedestalColor);
    pedestala = new DBox(0, hpedestal / 2. + tbase / 2.,
                         -d - 2 * gap - wpedestal / 2., wpedestal, hpedestal,
                         wpedestal, pedestalColor);
    axlea = new DCylinder(0, tbase / 2. + 0.9 * hpedestal,
                          -0.75 * d - 1.25 * gap, r, 1.5 * gap + d / 2.,
                          Color.yellow);
    axlea.rotate(Math.PI / 2., 0, 0);
    pedestalb = new DBox(0, hpedestal / 2. + tbase / 2.,
                         d + 2 * gap + wpedestal / 2., wpedestal, hpedestal,
                         wpedestal, pedestalColor);
    axleb = new DCylinder(0, tbase / 2. + 0.9 * hpedestal,
                          0.75 * d + 1.25 * gap, r, 1.5 * gap + d / 2.,
                          Color.yellow);
    axleb.rotate(Math.PI / 2., 0, 0);

    stationaryPart = new DShapeGroup();
    stationaryPart.addShape(base);
    stationaryPart.addShape(pedestala);
    stationaryPart.addShape(axlea);
    stationaryPart.addShape(pedestalb);
    stationaryPart.addShape(axleb);

    bar1a = new DBox(0, -L1 / 2., -0.75 * d - gap, d, L1display, d / 2.,
                     Color.red);
    bar1b = new DBox(0, -L1 / 2., 0.75 * d + gap, d, L1display, d / 2.,
                     Color.red);
    axlec = new DCylinder(0, -L1, 0, r, 3 * gap + 2 * d, Color.yellow);
    axlec.rotate(Math.PI / 2., 0, 0);

    upperPart = new DShapeGroup();
    upperPart.addShape(bar1a);
    upperPart.addShape(bar1b);
    upperPart.addShape(axlec);
    upperPart.setXYZ(0, tbase / 2. + 0.9 * hpedestal, 0);

    bar2 = new DBox(0, -L2 / 2., 0, d, L2display, d, Color.red);
    lowerPart = new DShapeGroup();
    lowerPart.addShape(bar2);
    lowerPart.setXYZ(0, tbase / 2. + 0.9 * hpedestal - L1, 0);

    if (trace) {
      dataset = new Dataset3D();
      dataset.setConnected(true);
      dataset.setLineColor(new Color(0.9f, 0.5f, 0.5f, 0.5f));
      dataset.setMarkerShape(Dataset3D.NO_MARKER);
      dataset.clear();
    }

    panel3D.addDrawable3D(stationaryPart);
    panel3D.addDrawable3D(upperPart);
    panel3D.addDrawable3D(lowerPart);
    if (trace)
      panel3D.addDrawable3D(dataset);

    panel3D.setSceneXYZ( -1.3715e-2, -1.3755, -4.6819);
    panel3D.setScenePEuler( -1.3163, 0.0580, 2.3363);

    upperPart.setPEuler(0, ode.getPhi1(), 0);
    lowerPart.setXYZ(L1 * Math.sin(ode.getPhi1()),
                     tbase / 2. + 0.9 * hpedestal - L1 * Math.cos(ode.getPhi1()),
                     0);
    lowerPart.setPEuler(0, ode.getPhi2(), 0);
    if (trace)
      dataset.append(L1 * Math.sin(ode.getPhi1()) + L2 * Math.sin(ode.getPhi2()),
                     tbase / 2. + 0.9 * hpedestal - L1 * Math.cos(ode.getPhi1()) -
                     L2 * Math.cos(ode.getPhi2()), 0);

    panel3D.repaint();
  }

  /**
   *
   */
  public void initializeAnimation() {
    //initial conditions
    phi1 = Math.toRadians(slPhi1.getValue());
    phi2 = Math.toRadians(slPhi2.getValue());
    w1 = Math.toRadians(slw1.getValue());
    w2 = Math.toRadians(slw2.getValue());
    L1 = 0.5;
    L2 = 1.0;
    t = 0;
    dt = 0.025;
    trace = true;

    //initialization of ODESolver
    odeAdaptiveSolver.initialize(dt);
    odeAdaptiveSolver.setTolerance(1e-9);
    odeAdaptiveSolver.setStepSize(dt);

    ode.setPhi1(phi1);
    ode.setPhi2(phi2);
    ode.setW1(w1);
    ode.setW2(w2);
    ode.setL1(L1);
    ode.setL2(L2);
    ode.setT(t);

    create3DScene();

    panel3D.repaint();
  }

  public void updateScene() {
    phi1 = Math.toRadians(slPhi1.getValue());
    phi2 = Math.toRadians(slPhi2.getValue());
    w1 = Math.toRadians(slw1.getValue());
    w2 = Math.toRadians(slw2.getValue());
    L1 = 0.5;
    L2 = 1.0;
    t = 0;
    dt = 0.025;
    trace = true;

    if (trace) {
      dataset.clear();
    }

    //initialization of ODESolver
    odeAdaptiveSolver.initialize(dt);
    odeAdaptiveSolver.setTolerance(1e-9);
    odeAdaptiveSolver.setStepSize(dt);

    ode.setPhi1(phi1);
    ode.setPhi2(phi2);
    ode.setW1(w1);
    ode.setW2(w2);
    ode.setL1(L1);
    ode.setL2(L2);
    ode.setT(t);

    upperPart.setPEuler(0, ode.getPhi1(), 0);
    lowerPart.setXYZ(L1 * Math.sin(ode.getPhi1()),
                     tbase / 2. + 0.9 * hpedestal - L1 * Math.cos(ode.getPhi1()),
                     0);
    lowerPart.setPEuler(0, ode.getPhi2(), 0);
    panel3D.repaint();
  }

  /**
   *
   */
  public void stepAnimation() {
    //execute one step of simulation
    odeAdaptiveSolver.step();

    //set the state of the pendulum
    upperPart.setPEuler(0, ode.getPhi1(), 0);
    lowerPart.setXYZ(L1 * Math.sin(ode.getPhi1()),
                     tbase / 2. + 0.9 * hpedestal - L1 * Math.cos(ode.getPhi1()),
                     0);
    lowerPart.setPEuler(0, ode.getPhi2(), 0);
    if (trace)
      dataset.append(L1 * Math.sin(ode.getPhi1()) + L2 * Math.sin(ode.getPhi2()),
                     tbase / 2. + 0.9 * hpedestal - L1 * Math.cos(ode.getPhi1()) -
                     L2 * Math.cos(ode.getPhi2()), 0);

    panel3D.repaint();
  }

  /**
   *
   */
  public void startAnimation() {
    timer.start();
  }

  /**
   *
   */
  public void stopAnimation() {
    timer.stop();
  }

  public void stateChanged(ChangeEvent e) {
    JSlider source = (JSlider) e.getSource();

    if (source.equals(slPhi1)) {
      slPhi1.setBorder(BorderFactory.createTitledBorder("\u03D5_1 = " +
          slPhi1.getValue() + " "));
      updateScene();
    }
    else if (source.equals(slPhi2)) {
      slPhi2.setBorder(BorderFactory.createTitledBorder("\u03D5_2 = " +
          slPhi2.getValue() + " "));
      updateScene();
    }
    else if (source.equals(slw1)) {
      slw1.setBorder(BorderFactory.createTitledBorder("d\u03D5_1 / dt = " +
          slw1.getValue() + " /s"));
      updateScene();
    }
    else if (source.equals(slw2)) {
      slw2.setBorder(BorderFactory.createTitledBorder("d\u03D5_2 / dt = " +
          slw2.getValue() + " /s"));
      updateScene();
    }

  }

}
