package org.opensourcephysics.tuleja.energydisplay;

import org.opensourcephysics.numerics.*;
/**
 * <p>Title: Energy display</p>
 * <p>Description: One dimensional motion of a particle with energy E in a potential</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Slavomir Tuleja
 */

public class EnergyODE implements ODE{
  Function U;
  Function Up;
  double E=1;
  int initialDirection=1;

  double[] state = new double[]{0,1,0};//{x,xDot,t}
  double m=1;

  public EnergyODE(Function U) {
    this.U=U;
    Up = new PotentialEnergyDerivative(U);
  }

  public double[] getState(){
    return state;
  }

  public void getRate(double[] state, double[] rates){
    rates[0]=state[1];
    rates[1]=-1.0/m * Up.evaluate(state[0]);
    rates[2]=1;
  }

  public void setInitialConditions(double x, double E){
    state[0]=x;
    double vel2 = (2.0/m)*(E-U.evaluate(x));
    if(vel2>=0) state[1]=initialDirection*Math.sqrt(vel2);
      else state[1]=0;
    state[2]=0;
    this.E=E;
  }

  public double getX(){
    return state[0];
  }

  public double getXDot(){
    return state[1];
  }

  public double getT(){
    return state[2];
  }

  public double getE(){
    return E;
  }

  /**
   * Returns maximim value of xDot for given potential function.
   * In this preliminary version we suppose the potential energy minimum has 0 value.
   * @return maximum value of xDot
   */
  public double getXDotMax(){
    //to be generalized later ...
    return Math.sqrt(2*5/m);
  }

  public void changeInitialDirection(){
    initialDirection*=-1;
  }

  public int getInitialDirection(){
    return initialDirection;
  }
}