package org.opensourcephysics.tuleja.energydisplay;
import java.util.*;
import javax.swing.*;
import java.awt.*;


public class PEStep implements Potential{
  Object hotPointList;
  int[] connectionType;
  int n;

  /**
   * Constructor.
   * @param x array of x-coordinates of hot points.
   * @param y array of y-coordinates of hot points.
   * @param connectionType array of connection types for pairs of adjacent hot points, where 0 - straight connection, 1 - _| connection , 2 - |_ connection
   */
  public PEStep(double[] x, double[] y, int[] connectionType) {
    int n;
    if(x.length < y.length){
      n = x.length;
    }
    else n=y.length;
    this.n=n;
    this.connectionType=connectionType;

    hotPointList = new LinkedList();
    for (int i = 0; i < n; i++) {
      ( (java.util.List) hotPointList).add(new HotPoint(x[i], y[i]));
    }

  }


  public void draw(Graphics2D g2, Scale sc, boolean showHotPoints, boolean editPotential){
    float[] dashed = {
        2.0f, 2.0f};
    float[] solid = {
        12.0f, 0.0f};

    BasicStroke bsDashed = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
                                           BasicStroke.JOIN_MITER, 1.0f, dashed,
                                           0.0f);
    BasicStroke bsSolid = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
                                          BasicStroke.JOIN_MITER, 1.0f, solid,
                                          0.0f);
    BasicStroke bsBold = new BasicStroke(2.0f);
    BasicStroke bsDoubleBold = new BasicStroke(4.0f);

    //potential function
    g2.setStroke(bsBold);
    ListIterator hpit = ( (java.util.List) hotPointList).listIterator();

    int k=0, i1, j1, i2, j2;
    HotPoint hp1, hp2;
    if (hpit.hasNext()) {
      hp1 = (HotPoint) hpit.next();
      i1 = sc.xToi(hp1.getX());
      j1 = sc.yToj(hp1.getY());
      g2.drawLine(sc.xToi(-2), sc.yToj(5), i1, j1);

      while (hpit.hasNext()) {
        hp2 = (HotPoint) hpit.next();
        i2 = sc.xToi(hp2.getX());
        j2 = sc.yToj(hp2.getY());
        if(connectionType[k]==1) {
          g2.drawLine(i1, j1, i2, j1);
          g2.drawLine(i2, j1, i2, j2);

        }
        else if(connectionType[k]==2){
          g2.drawLine(i1, j1, i1, j2);
          g2.drawLine(i1, j2, i2, j2);
        }
        else {
          g2.drawLine(i1, j1, i2, j2);
        }

        k++;
        hp1 = hp2;
        i1=i2;
        j1=j2;
      }
      g2.drawLine(i1, j1, sc.xToi(2), sc.yToj(5));

    }

    g2.setStroke(bsSolid);

    //hot points on the potential function
    if (showHotPoints) {
      hpit = ( (java.util.List) hotPointList).listIterator();

      while (hpit.hasNext()) {
        HotPoint hp = (HotPoint) hpit.next();
        int i = sc.xToi(hp.getX());
        int j = sc.yToj(hp.getY());
        g2.setColor(Color.red);
        g2.fillRect(i - 8, j - 8, 16, 16);
        g2.setColor(Color.black);
        g2.drawRect(i - 8, j - 8, 16, 16);
      }
    }
  }

  public Object getHotPointList(){
    return hotPointList;
  }

  public void setHotPoint(int i, double x, double y){
    HotPoint hp, hp1, hp2;
    if(i==1){
      hp=(HotPoint)(((java.util.List)hotPointList).get(0));
      hp2=(HotPoint)(((java.util.List)hotPointList).get(1));
      if((x<hp2.getX()-0.02)&&(x>-2+0.02)) hp.x=x;
      hp.y=y;
    }
    else if(i==n){
      hp=(HotPoint)(((java.util.List)hotPointList).get(n-1));
      hp1=(HotPoint)(((java.util.List)hotPointList).get(n-2));
      if((x<2-0.02)&&(x>hp1.getX()+0.02)) hp.x=x;
      hp.y=y;
    }
    else {
      hp=(HotPoint)(((java.util.List)hotPointList).get(i-1));
      hp1=(HotPoint)(((java.util.List)hotPointList).get(i-2));
      hp2=(HotPoint)(((java.util.List)hotPointList).get(i));

      if((x<hp2.getX()-0.02)&&(x>hp1.getX()+0.02)) hp.x=x;
      hp.y=y;
    }
  }

  private double interpolate(double x, double x1, double y1, double x2, double y2){
    if(Math.abs(x2-x1)>1e-6) return y1+(y2-y1)*(x-x1)/(x2-x1);
    else {
      if (y1 > y2)
        return y1;
      else return y2;
    }
  }

  public double evaluate (double x){
    double y=0;
    ListIterator hpit = ( (java.util.List) hotPointList).listIterator();

    int k=0;
    HotPoint hp1, hp2;
    if (hpit.hasNext()) {
      hp1 = (HotPoint) hpit.next();
      if(x<=hp1.getX()) y=interpolate(x,-2,5,hp1.getX(),hp1.getY());
      while (hpit.hasNext()) {
        hp2 = (HotPoint) hpit.next();
        if((hp1.getX()<x)&&(x<=hp2.getX())){
          if ( connectionType[k]==0) {
            y = interpolate(x, hp1.getX(), hp1.getY(), hp2.getX(), hp2.getY());
          }
          else if ( connectionType[k]==2) {
            double xtemp=hp1.getX()+0.02;
            if(x<xtemp) y=interpolate(x,hp1.getX(),hp1.getY(),xtemp,hp2.getY());
            else y=hp2.getY();
          }
          else {
            double xtemp=hp2.getX()-0.02;
            if(x>xtemp) y=interpolate(x,xtemp,hp1.getY(),hp2.getX(),hp2.getY());
            else y=hp1.getY();
          }
        }

        k++;
        hp1 = hp2;
      }
      if(hp1.getX()<x){
        y=interpolate(x,hp1.getX(),hp1.getY(),2,5);
      }
    }

    return y;
  }

}