package org.opensourcephysics.tuleja.feynmanplate;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import org.opensourcephysics.displayejs.*;
import java.text.*;
import java.io.*;
import javax.swing.event.*;
import javax.swing.text.html.*;
import java.awt.image.*;

/**
 * <p>Title: JFeynmanPlate</p>
 * <p>Description: JApplet class. This is the main class of the project. It creates
 * the GUI and defines button click, menu click, and mouse actions.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * @author Alexander Tomori, Boris Gazovic, Slavomir Tuleja
 * @version 0.01 of February 27, 2004
 */
public class JFeynmanPlate
    extends JApplet
    implements Runnable, ActionListener, ChangeListener, HyperlinkListener {

  Container cp;
  JPanel pnlSouth, pnlSliders, pnlButtons, pnlHelpButtons;
  JButton btnStartStopResume, btnReset, btnStepForward, btnStepBack,
      btnBackToSimulation, btnHelp, btnAbout, btnTitlePageContinue;
  ButtonGroup bgBody, bgFrame, bgProjection;
  JMenuBar menuBar;
  JMenu menuRigidBody, menuFrame, menuShow, menuProjection, menuHelp;
  JRadioButtonMenuItem rbMenuItemPlate, rbMenuItemDisc,
      rbMenuItemElementaryPlate, rbMenuItemSphere, rbMenuItemCuboid,
      rbMenuItemBodyFixedFrame, rbMenuItemLaboratoryFrame,
      rbMenuItemXYProjection, rbMenuItemXZProjection, rbMenuItemYZProjection,
      rbMenuItem3DProjection, rbMenuItem3DProjectionPerspective;
  JCheckBoxMenuItem chbMenuItemBody, chbMenuItemBodyFixedAxes,
      chbMenuItemLabAxes, chbMenuItemOmegaL, chbMenuItemTraces;
  JMenuItem menuItemHelp, menuItemAbout;
  JSlider slTheta, slPhi, slPsi, slThetaDot, slPhiDot, slPsiDot;

  JEditorPane editorPane;
  JScrollPane scrollPane;

  Thread thread = null;

  DrawingPanel3D panel;
  RigidBody body;

  DecimalFormat format = new DecimalFormat();
  boolean running = false;

  Image imStart, imPause, imForward, imBack;
  ImageIcon iiStart, iiPause, iiForward, iiBack;

  //help page
  java.net.URL helpURL = JFeynmanPlate.class.getResource(
      "/org/opensourcephysics/tuleja/feynmanplate/html/help.html");
  //title page
  java.net.URL titleURL = JFeynmanPlate.class.getResource(
      "/org/opensourcephysics/tuleja/feynmanplate/html/title.html");
  //about page
  java.net.URL aboutURL = JFeynmanPlate.class.getResource(
      "/org/opensourcephysics/tuleja/feynmanplate/html/about.html");

  //captions of buttons and all texts used in GUI
  String paramStartStopResume = new String("Start/Pause");
  String paramReset = new String("Reset");
  String paramStepForward = new String("Step forward");
  String paramStepBack = new String("Step back");
  String paramBackToSimulation = new String("Simulation...");
  //captions of menus
  String paramMenuRigidBody = new String("Rigid Body");
  String paramMenuFrame = new String("Reference Frame");
  String paramMenuShow = new String("Show");
  String paramMenuProjection = new String("Projection");
  String paramMenuHelp = new String("Help");
  //captions of Rigid Body menu items
  String paramMiPlate = new String("Plate");
  String paramMiDisc = new String("Disc");
  String paramMiElementaryPlate = new String("Elementary Plate");
  String paramMiSphere = new String("Sphere");
  String paramMiCuboid = new String("Cuboid");
  //captions of Reference Frame menu items
  String paramMiBodyFixedFrame = new String("Body-fixed frame");
  String paramMiLaboratoryFrame = new String("Laboratory frame");
  //captions of Projection menu items
  String paramMiXYProjection = new String("floor");
  String paramMiXZProjection = new String("side");
  String paramMiYZProjection = new String("frontal");
  String paramMi3DProjection = new String("3D without perspective");
  String paramMi3DProjectionPerspective = new String("3D with perspective");
  //captions of Show menu items
  String paramMiBody = new String("Rigid Body");
  String paramMiBodyFixedAxes = new String("Body-fixed axes");
  String paramMiLabAxes = new String("Laboratory axes");
  String paramMiOmegaL = new String("Vectors \u03C9 and L");
  String paramMiTraces = new String("Vector traces");
  //captions of Help items
  String paramMiHelp = new String("Help..."); //also used for btnHelp in help layout
  String paramMiAbout = new String("About..."); //also used for btnAbout in help layout
  //
  String paramPanel3DLabTitle = new String(
      "Virtual 3D world - Laboratory frame");
  String paramPanel3DBodyTitle = new String(
      "Virtual 3D world - Body-fixed frame");
  String paramSlidersTitle = new String("Initial conditions for simulation:");

  String paramNext = new String("Next");

  public static void main(String[] args) {
    JFrame fr = new JFrame("JDoublePendulum");
    fr.setSize(358, 610);
    JFeynmanPlate applet = new JFeynmanPlate();
    applet.initApplet();
    applet.start();
    fr.getContentPane().add(applet, BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }

  public void init() {
    //texts will be translated only in case of different parameter value from "en"
    if (!getParameter("language").equals("en"))
      translateTexts();
    initApplet();
  }

  /** Returns an ImageIcon, or null if the path was invalid. */
  protected ImageIcon createImageIcon(String path) {
    java.net.URL imgURL = JFeynmanPlate.class.getResource(path);
    if (imgURL != null) {
      return new ImageIcon(imgURL);
    }
    else {
      System.err.println("Couldn't find file: " + path);
      return null;
    }
  }

  public void initApplet() {
    //creation of GUI components
    //setting Java look and feel
    try {
      UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
    }
    catch (Exception e) {
      System.err.println("Couldn't use the cross-platform " + "look and feel: " +
                         e);
    }

    //loading images

    String imagePath = "/org/opensourcephysics/tuleja/images/";
    iiStart = createImageIcon(imagePath + "play16.gif");
    iiPause = createImageIcon(imagePath + "pause16.gif");
    iiForward = createImageIcon(imagePath + "stepforward16.gif");
    iiBack = createImageIcon(imagePath + "stepback16.gif");

    //buttons
    btnStartStopResume = new JButton(iiStart);
    btnStartStopResume.addActionListener(this);
    btnStartStopResume.setToolTipText(paramStartStopResume);
    btnStartStopResume.setActionCommand("start");
    btnReset = new JButton(paramReset);
    //btnReset.setToolTipText(paramReset);
    btnReset.addActionListener(this);
    btnReset.setActionCommand("reset");
    btnStepForward = new JButton(iiForward);
    btnStepForward.addActionListener(this);
    btnStepForward.setActionCommand("stepForward");
    btnStepForward.setToolTipText(paramStepForward);
    btnStepBack = new JButton(iiBack);
    btnStepBack.addActionListener(this);
    btnStepBack.setActionCommand("stepBack");
    btnStepBack.setToolTipText(paramStepBack);

    btnBackToSimulation = new JButton(paramBackToSimulation);
    btnBackToSimulation.setActionCommand("backToSimulation");
    btnBackToSimulation.addActionListener(this);
    btnHelp = new JButton(paramMiHelp);
    btnHelp.setActionCommand("help");
    btnHelp.addActionListener(this);
    btnAbout = new JButton(paramMiAbout);
    btnAbout.setActionCommand("about");
    btnAbout.addActionListener(this);
    btnTitlePageContinue = new JButton(paramNext);
    btnTitlePageContinue.setActionCommand("next");
    btnTitlePageContinue.addActionListener(this);
    //buttons btnHelp and btnAbout are used in help layout and share the same
    //captions and action commands with menu items menuItemHelp and menuItemAhout
    //they have the same function

    //menu
    menuBar = new JMenuBar();
    menuRigidBody = new JMenu(paramMenuRigidBody);
    menuFrame = new JMenu(paramMenuFrame);
    menuShow = new JMenu(paramMenuShow);
    menuProjection = new JMenu(paramMenuProjection);
    menuHelp = new JMenu(paramMenuHelp);

    menuBar.add(menuRigidBody);
    menuBar.add(menuFrame);
    menuBar.add(menuShow);
    menuBar.add(menuProjection);
    menuBar.add(menuHelp);

    rbMenuItemPlate = new JRadioButtonMenuItem(paramMiPlate, true);
    rbMenuItemPlate.setActionCommand("plate");
    rbMenuItemPlate.addActionListener(this);
    rbMenuItemDisc = new JRadioButtonMenuItem(paramMiDisc, false);
    rbMenuItemDisc.setActionCommand("disc");
    rbMenuItemDisc.addActionListener(this);
    rbMenuItemElementaryPlate = new JRadioButtonMenuItem(paramMiElementaryPlate, false);
    rbMenuItemElementaryPlate.setActionCommand("elementaryPlate");
    rbMenuItemElementaryPlate.addActionListener(this);
    rbMenuItemSphere = new JRadioButtonMenuItem(paramMiSphere, false);
    rbMenuItemSphere.setActionCommand("sphere");
    rbMenuItemSphere.addActionListener(this);
    rbMenuItemCuboid = new JRadioButtonMenuItem(paramMiCuboid, false);
    rbMenuItemCuboid.setActionCommand("cuboid");
    rbMenuItemCuboid.addActionListener(this);
    rbMenuItemXYProjection = new JRadioButtonMenuItem(paramMiXYProjection, false);
    rbMenuItemXYProjection.setActionCommand("xyProjection");
    rbMenuItemXYProjection.addActionListener(this);
    rbMenuItemXZProjection = new JRadioButtonMenuItem(paramMiXZProjection, false);
    rbMenuItemXZProjection.setActionCommand("xzProjection");
    rbMenuItemXZProjection.addActionListener(this);
    rbMenuItemYZProjection = new JRadioButtonMenuItem(paramMiYZProjection, false);
    rbMenuItemYZProjection.setActionCommand("yzProjection");
    rbMenuItemYZProjection.addActionListener(this);
    rbMenuItem3DProjection = new JRadioButtonMenuItem(paramMi3DProjection, false);
    rbMenuItem3DProjection.setActionCommand("3DProjection");
    rbMenuItem3DProjection.addActionListener(this);
    rbMenuItem3DProjectionPerspective = new JRadioButtonMenuItem(
        paramMi3DProjectionPerspective, true);
    rbMenuItem3DProjectionPerspective.setActionCommand(
        "3DProjectionPerspective");
    rbMenuItem3DProjectionPerspective.addActionListener(this);

    bgProjection = new ButtonGroup();
    bgProjection.add(rbMenuItemXYProjection);
    bgProjection.add(rbMenuItemXZProjection);
    bgProjection.add(rbMenuItemYZProjection);
    bgProjection.add(rbMenuItem3DProjection);
    bgProjection.add(rbMenuItem3DProjectionPerspective);

    menuProjection.add(rbMenuItem3DProjectionPerspective);
    menuProjection.add(rbMenuItem3DProjection);
    menuProjection.add(rbMenuItemXYProjection);
    menuProjection.add(rbMenuItemXZProjection);
    menuProjection.add(rbMenuItemYZProjection);

    bgBody = new ButtonGroup();
    bgBody.add(rbMenuItemPlate);
    bgBody.add(rbMenuItemDisc);
    bgBody.add(rbMenuItemElementaryPlate);
    bgBody.add(rbMenuItemSphere);
    bgBody.add(rbMenuItemCuboid);

    menuRigidBody.add(rbMenuItemPlate);
    menuRigidBody.add(rbMenuItemDisc);
    menuRigidBody.add(rbMenuItemElementaryPlate);
    menuRigidBody.add(rbMenuItemSphere);
    menuRigidBody.add(rbMenuItemCuboid);

    rbMenuItemBodyFixedFrame = new JRadioButtonMenuItem(paramMiBodyFixedFrame, false);
    rbMenuItemBodyFixedFrame.setActionCommand("bodyFixedFrame");
    rbMenuItemBodyFixedFrame.addActionListener(this);
    rbMenuItemLaboratoryFrame = new JRadioButtonMenuItem(paramMiLaboratoryFrame, true);
    rbMenuItemLaboratoryFrame.setActionCommand("laboratoryFrame");
    rbMenuItemLaboratoryFrame.addActionListener(this);

    bgFrame = new ButtonGroup();
    bgFrame.add(rbMenuItemLaboratoryFrame);
    bgFrame.add(rbMenuItemBodyFixedFrame);

    menuFrame.add(rbMenuItemLaboratoryFrame);
    menuFrame.add(rbMenuItemBodyFixedFrame);

    chbMenuItemBody = new JCheckBoxMenuItem(paramMiBody, true);
    chbMenuItemBody.setActionCommand("rigidBody");
    chbMenuItemBody.addActionListener(this);
    chbMenuItemBodyFixedAxes = new JCheckBoxMenuItem(paramMiBodyFixedAxes, false);
    chbMenuItemBodyFixedAxes.setActionCommand("bodyFixedAxes");
    chbMenuItemBodyFixedAxes.addActionListener(this);
    chbMenuItemLabAxes = new JCheckBoxMenuItem(paramMiLabAxes, false);
    chbMenuItemLabAxes.setActionCommand("laboratoryAxes");
    chbMenuItemLabAxes.addActionListener(this);
    paramMiOmegaL = paramMiOmegaL.replaceAll("omega", "\u03c9");
    chbMenuItemOmegaL = new JCheckBoxMenuItem(paramMiOmegaL, false);
    chbMenuItemOmegaL.setActionCommand("vectorsOmegaL");
    chbMenuItemOmegaL.addActionListener(this);
    chbMenuItemTraces = new JCheckBoxMenuItem(paramMiTraces, false);
    chbMenuItemTraces.setActionCommand("vectorTraces");
    chbMenuItemTraces.addActionListener(this);

    menuShow.add(chbMenuItemBody);
    menuShow.add(chbMenuItemBodyFixedAxes);
    menuShow.add(chbMenuItemLabAxes);
    menuShow.add(chbMenuItemOmegaL);
    menuShow.add(chbMenuItemTraces);

    menuItemHelp = new JMenuItem(paramMiHelp);
    menuItemHelp.setActionCommand("help");
    menuItemHelp.addActionListener(this);
    menuItemAbout = new JMenuItem(paramMiAbout);
    menuItemAbout.setActionCommand("about");
    menuItemAbout.addActionListener(this);

    menuHelp.add(menuItemHelp);
    menuHelp.add(menuItemAbout);

    setJMenuBar(menuBar);

    //sliders
    slTheta = new JSlider(JSlider.HORIZONTAL, 1, 179, 15);
    slTheta.setBorder(BorderFactory.createTitledBorder("\u03B8 = " +
        slTheta.getValue() + ""));
    slTheta.addChangeListener(this);

    slPhi = new JSlider(JSlider.HORIZONTAL, 0, 360, 45);
    slPhi.setBorder(BorderFactory.createTitledBorder("\u03D5 = " +
        slPhi.getValue() + ""));
    slPhi.addChangeListener(this);

    slPsi = new JSlider(JSlider.HORIZONTAL, 0, 360, 10);
    slPsi.setBorder(BorderFactory.createTitledBorder("\u03C8 = " +
        slPsi.getValue() + ""));
    slPsi.addChangeListener(this);

    slThetaDot = new JSlider(JSlider.HORIZONTAL, -200, 200, 50);
    slThetaDot.setBorder(BorderFactory.createTitledBorder("d\u03B8/dt = " +
        slThetaDot.getValue() + "/s"));
    slThetaDot.addChangeListener(this);

    slPhiDot = new JSlider(JSlider.HORIZONTAL, -200, 200, 0);
    slPhiDot.setBorder(BorderFactory.createTitledBorder("d\u03D5/dt = " +
        slPhiDot.getValue() + "/s"));
    slPhiDot.addChangeListener(this);

    slPsiDot = new JSlider(JSlider.HORIZONTAL, -500, 500, 450);
    slPsiDot.setBorder(BorderFactory.createTitledBorder("d\u03C8/dt = " +
        slPsiDot.getValue() + "/s"));
    slPsiDot.addChangeListener(this);

    //Plate object
    body = new Plate();

    panel = new DrawingPanel3D();
    panel.setPreferredMinMax( -1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
    panel.addDrawable(body);
    panel.setAlphaAndBeta(Math.toRadians(30), Math.toRadians(20));
    panel.setDisplayMode(DrawingPanel3D.DISPLAY_PERSPECTIVE);
    panel.setDecorationType(DrawingPanel3D.DECORATION_NONE);
    panel.repaint();
    panel.setBorder(BorderFactory.createTitledBorder(paramPanel3DLabTitle));

    pnlSliders = new JPanel();
    pnlSliders.setLayout(new GridLayout(2, 3));
    pnlSliders.add(slTheta);
    pnlSliders.add(slPhi);
    pnlSliders.add(slPsi);
    pnlSliders.add(slThetaDot);
    pnlSliders.add(slPhiDot);
    pnlSliders.add(slPsiDot);
    pnlSliders.setBorder(BorderFactory.createTitledBorder(paramSlidersTitle));

    pnlButtons = new JPanel();
    pnlButtons.setLayout(new GridLayout(1, 4));
    pnlButtons.add(btnStartStopResume);
    pnlButtons.add(btnStepBack);
    pnlButtons.add(btnStepForward);
    pnlButtons.add(btnReset);
    pnlButtons.setBorder(BorderFactory.createTitledBorder("t = 0 s"));

    pnlSouth = new JPanel();
    pnlSouth.setLayout(new BoxLayout(pnlSouth, BoxLayout.Y_AXIS));
    pnlSouth.add(pnlSliders);
    pnlSouth.add(pnlButtons);

    cp = getContentPane();

    pnlHelpButtons = new JPanel(new GridLayout()); //panel which contains buttons used in help layout

    editorPane = new JEditorPane();
    editorPane.setEditable(false);
    editorPane.addHyperlinkListener(this);

    //System.out.println(getCodeBase());
    //helpURL = JFeynmanPlate.class.getResource(getCodeBase().toString().concat("help.html"));

    if (titleURL != null) {
      try {
        editorPane.setPage(titleURL);
      }
      catch (IOException e) {
        System.err.println("Attempted to read a bad URL: " + titleURL);
      }
    }
    else {
      System.err.println("Couldn't find file: title.html");
    }

    menuBar.setVisible(false);

    createTitlePageLayout();
  }

  public void createSimulationLayout() {
    //components added to cp
    cp.removeAll();
    cp.setLayout(new BorderLayout());
    cp.add(pnlSouth, BorderLayout.SOUTH);
    cp.add(panel, BorderLayout.CENTER);
    cp.validate();
    cp.repaint();
  }

  public void createHelpLayout() {
    //components added to cp
    cp.removeAll();
    cp.setLayout(new BorderLayout());

    //Put the editor pane in a scroll pane.
    scrollPane = new JScrollPane(editorPane);
    editorPane.setBorder(BorderFactory.createEmptyBorder(8, 8, 8, 8));
    scrollPane.setVerticalScrollBarPolicy(JScrollPane.
                                          VERTICAL_SCROLLBAR_AS_NEEDED);
    scrollPane.setHorizontalScrollBarPolicy(JScrollPane.
                                            HORIZONTAL_SCROLLBAR_AS_NEEDED);

    cp.add(scrollPane, BorderLayout.CENTER);
    cp.add(pnlHelpButtons, BorderLayout.SOUTH);
    pnlHelpButtons.add(btnBackToSimulation);
    pnlHelpButtons.add(btnHelp);
    pnlHelpButtons.add(btnAbout);
    cp.validate();
    cp.repaint();
  }

  public void createTitlePageLayout() {
    cp.setLayout(new BorderLayout());
    editorPane.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
    cp.add(editorPane, BorderLayout.CENTER);
    cp.add(btnTitlePageContinue, BorderLayout.SOUTH);

    cp.validate();
    cp.repaint();
  }

  public void start() {
    body.setInitialConditions(Math.toRadians( (double) slTheta.getValue()),
                              Math.toRadians( (double) slPhi.getValue()),
                              Math.toRadians( (double) slPsi.getValue()),
                              Math.toRadians( (double) slThetaDot.getValue()),
                              Math.toRadians( (double) slPhiDot.getValue()),
                              Math.toRadians( (double) slPsiDot.getValue()));
    body.transform(Math.toRadians( (double) slTheta.getValue()),
                   Math.toRadians( (double) slPhi.getValue()),
                   Math.toRadians( (double) slPsi.getValue()));
    panel.repaint();
  }

  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("start")) {
      btnStartStopResume.setIcon(iiPause);
      btnStartStopResume.setActionCommand("stop");
      btnReset.setEnabled(false);
      btnStepForward.setEnabled(false);
      btnStepBack.setEnabled(false);
      slTheta.setEnabled(false);
      slThetaDot.setEnabled(false);
      slPhi.setEnabled(false);
      slPhiDot.setEnabled(false);
      slPsi.setEnabled(false);
      slPsiDot.setEnabled(false);

      menuRigidBody.setEnabled(false);
      menuFrame.setEnabled(false);
      menuShow.setEnabled(false);
      menuProjection.setEnabled(false);
      menuHelp.setEnabled(false);

      running = true;
      thread = new Thread(this);
      thread.start();
    }
    else if (e.getActionCommand().equals("stop")) {
      btnStartStopResume.setIcon(iiStart);
      btnStartStopResume.setActionCommand("resume");
      btnReset.setEnabled(true);
      btnStepForward.setEnabled(true);
      btnStepBack.setEnabled(true);

      menuFrame.setEnabled(true);
      menuShow.setEnabled(true);
      menuProjection.setEnabled(true);
      menuHelp.setEnabled(true);

      running = false;
    }
    else if (e.getActionCommand().equals("resume")) {
      btnStartStopResume.setActionCommand("stop");
      btnStartStopResume.setIcon(iiPause);
      btnReset.setEnabled(false);
      btnStepForward.setEnabled(false);
      btnStepBack.setEnabled(false);

      menuFrame.setEnabled(false);
      menuShow.setEnabled(false);
      menuProjection.setEnabled(false);
      menuHelp.setEnabled(false);

      running = true;
      thread = new Thread(this);
      thread.start();
    }
    else if (e.getActionCommand().equals("reset")) {
      btnStartStopResume.setActionCommand("start");
      btnStartStopResume.setIcon(iiStart);
      btnReset.setEnabled(true);
      btnStepForward.setEnabled(true);
      btnStepBack.setEnabled(true);
      slTheta.setEnabled(true);
      slThetaDot.setEnabled(true);
      slPhi.setEnabled(true);
      slPhiDot.setEnabled(true);
      slPsi.setEnabled(true);
      slPsiDot.setEnabled(true);

      menuRigidBody.setEnabled(true);

      start();
      pnlButtons.setBorder(BorderFactory.createTitledBorder(format.format(body.
          getT()).concat(" s")));
    }
    else if (e.getActionCommand().equals("stepForward")) {
      body.doStep();
      panel.repaint();
      pnlButtons.setBorder(BorderFactory.createTitledBorder(format.format(body.
          getT()).concat(" s")));
    }
    else if (e.getActionCommand().equals("stepBack")) {
      body.solver.setStepSize( -body.solver.getStepSize());
      body.doStep();
      panel.repaint();
      pnlButtons.setBorder(BorderFactory.createTitledBorder(format.format(body.
          getT()).concat(" s")));
      body.solver.setStepSize( -body.solver.getStepSize());
    }
    else if (e.getActionCommand().equals("plate")) {
      panel.removeDrawable(body);
      body = new Plate();
      applyMenuSettings(body);
      panel.addDrawable(body);
      start();
    }
    else if (e.getActionCommand().equals("disc")) {
      panel.removeDrawable(body);
      body = new Disc();
      applyMenuSettings(body);
      panel.addDrawable(body);
      start();
    }
    else if (e.getActionCommand().equals("elementaryPlate")) {
      panel.removeDrawable(body);
      body = new ElementaryPlate();
      applyMenuSettings(body);
      panel.addDrawable(body);
      start();
    }
    else if (e.getActionCommand().equals("sphere")) {
      panel.removeDrawable(body);
      body = new Sphere();
      applyMenuSettings(body);
      panel.addDrawable(body);
      start();
    }
    else if (e.getActionCommand().equals("cuboid")) {
      panel.removeDrawable(body);
      body = new Cuboid();
      applyMenuSettings(body);
      panel.addDrawable(body);
      start();
    }
    else if (e.getActionCommand().equals("bodyFixedFrame")) {
      body.isBodyFixedFrame = true;
      body.clearTraces();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.setBorder(BorderFactory.createTitledBorder(paramPanel3DBodyTitle));
      panel.repaint();
    }
    else if (e.getActionCommand().equals("laboratoryFrame")) {
      body.isBodyFixedFrame = false;
      body.clearTraces();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.setBorder(BorderFactory.createTitledBorder(paramPanel3DLabTitle));
      panel.repaint();
    }
    else if (e.getActionCommand().equals("rigidBody")) {
      body.showRigidBody = chbMenuItemBody.getState();
      body.setVisibility();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.repaint();
    }
    else if (e.getActionCommand().equals("bodyFixedAxes")) {
      body.showBodyFixedAxes = chbMenuItemBodyFixedAxes.getState();
      body.setVisibility();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.repaint();
    }
    else if (e.getActionCommand().equals("laboratoryAxes")) {
      body.showLaboratoryAxes = chbMenuItemLabAxes.getState();
      body.setVisibility();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.repaint();
    }
    else if (e.getActionCommand().equals("vectorsOmegaL")) {
      body.showOmegaL = chbMenuItemOmegaL.getState();
      body.setVisibility();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.repaint();
    }
    else if (e.getActionCommand().equals("vectorTraces")) {
      body.showTraces = chbMenuItemTraces.getState();
      body.setVisibility();
      body.clearTraces();
      body.transform(body.getTheta(), body.getPhi(), body.getPsi());
      panel.repaint();
    }
    else if (e.getActionCommand().equals("xyProjection")) {
      panel.setDisplayMode(DrawingPanel3D.DISPLAY_PLANAR_XY);
      panel.repaint();
    }
    else if (e.getActionCommand().equals("xzProjection")) {
      panel.setDisplayMode(DrawingPanel3D.DISPLAY_PLANAR_XZ);
      panel.repaint();
    }
    else if (e.getActionCommand().equals("yzProjection")) {
      panel.setDisplayMode(DrawingPanel3D.DISPLAY_PLANAR_YZ);
      panel.repaint();
    }
    else if (e.getActionCommand().equals("3DProjection")) {
      panel.setDisplayMode(DrawingPanel3D.DISPLAY_NO_PERSPECTIVE);
      panel.repaint();
    }
    else if (e.getActionCommand().equals("3DProjectionPerspective")) {
      panel.setDisplayMode(DrawingPanel3D.DISPLAY_PERSPECTIVE);
      panel.repaint();
    }
    else if (e.getActionCommand().equals("help")) {
      createHelpLayout();
      menuBar.setVisible(false);
      if (helpURL != null) {
        try {
          editorPane.setPage(helpURL);
        }
        catch (IOException ex) {
          System.err.println("Attempted to read a bad URL: " + helpURL);
        }
      }
      else {
        System.err.println("Couldn't find file: " + helpURL);
      }
    }
    else if (e.getActionCommand().equals("about")) {
      createHelpLayout();
      menuBar.setVisible(false);
      if (aboutURL != null) {
        try {
          editorPane.setPage(aboutURL);
        }
        catch (IOException ex) {
          System.err.println("Attempted to read a bad URL: " + aboutURL);
        }
      }
      else {
        System.err.println("Couldn't find file: " + aboutURL);
      }
    }
    else if (e.getActionCommand().equals("backToSimulation")) {
      menuBar.setVisible(true);
      createSimulationLayout();
    }
    else if (e.getActionCommand().equals("next")) {
      menuBar.setVisible(true);
      createSimulationLayout();
    }
  }

  public void hyperlinkUpdate(HyperlinkEvent e) {
    if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
      JEditorPane pane = (JEditorPane) e.getSource();
      if (e instanceof HTMLFrameHyperlinkEvent) {
        HTMLFrameHyperlinkEvent evt = (HTMLFrameHyperlinkEvent) e;
        HTMLDocument doc = (HTMLDocument) pane.getDocument();
        doc.processHTMLFrameHyperlinkEvent(evt);
      }
      else {
        try {
          pane.setPage(e.getURL());
        }
        catch (Throwable t) {
          t.printStackTrace();
        }
      }
    }
  }

  /**
   * Useful after we create a new RigidBody object. It tells it what the menu
   * settings are, calls its setVisibility() method and repaints the panel.
   * @param body RigidBody
   */
  public void applyMenuSettings(RigidBody body) {
    body.showRigidBody = chbMenuItemBody.getState();
    body.showBodyFixedAxes = chbMenuItemBodyFixedAxes.getState();
    body.showLaboratoryAxes = chbMenuItemLabAxes.getState();
    body.showOmegaL = chbMenuItemOmegaL.getState();
    body.showTraces = chbMenuItemTraces.getState();
    if (rbMenuItemBodyFixedFrame.isSelected())
      body.isBodyFixedFrame = true;
    else
      body.isBodyFixedFrame = false;
    body.setVisibility();
    panel.repaint();
  }

  public void stateChanged(ChangeEvent e) {
    JSlider source = (JSlider) e.getSource();

    if (source.equals(slTheta)) {
      slTheta.setBorder(BorderFactory.createTitledBorder("\u03B8 = " +
          slTheta.getValue() + ""));
    }
    else if (source.equals(slPhi)) {
      slPhi.setBorder(BorderFactory.createTitledBorder("\u03D5 = " +
          slPhi.getValue() + ""));
    }
    else if (source.equals(slPsi)) {
      slPsi.setBorder(BorderFactory.createTitledBorder("\u03C8 = " +
          slPsi.getValue() + ""));
    }
    else if (source.equals(slThetaDot)) {
      slThetaDot.setBorder(BorderFactory.createTitledBorder("d\u03B8/dt = " +
          slThetaDot.getValue() + "/s"));
    }
    else if (source.equals(slPhiDot)) {
      slPhiDot.setBorder(BorderFactory.createTitledBorder("d\u03D5/dt = " +
          slPhiDot.getValue() + "/s"));
    }
    else if (source.equals(slPsiDot)) {
      slPsiDot.setBorder(BorderFactory.createTitledBorder("d\u03C8/dt = " +
          slPsiDot.getValue() + "/s"));
    }

    body.setInitialConditions(Math.toRadians( (double) slTheta.getValue()),
                              Math.toRadians( (double) slPhi.getValue()),
                              Math.toRadians( (double) slPsi.getValue()),
                              Math.toRadians( (double) slThetaDot.getValue()),
                              Math.toRadians( (double) slPhiDot.getValue()),
                              Math.toRadians( (double) slPsiDot.getValue()));
    panel.repaint();
  }

  public void run() {
    double time;
    while (running) {
      body.doStep();
      panel.repaint();
      time = body.getT();
      pnlButtons.setBorder(BorderFactory.createTitledBorder(format.format(time).
          concat(" s")));

      try {
        thread.sleep(50);
      }
      catch (InterruptedException exception) {
        System.err.println("Thread interrupted...");
      }
    }

  }

  public void translateTexts() {
    //buttons
    paramStartStopResume = translateCaption("paramStartStopResume");

    paramStepForward = translateCaption("paramStepForward");
    paramStepBack = translateCaption("paramStepBack");

    paramBackToSimulation = translateCaption("paramBackToSimulation");
    //menus
    paramMenuRigidBody = translateCaption("paramMenuRigidBody");
    paramMenuFrame = translateCaption("paramMenuFrame");
    paramMenuShow = translateCaption("paramMenuShow");
    paramMenuProjection = translateCaption("paramMenuProjection");
    paramMenuHelp = translateCaption("paramMenuHelp");
    //menu Rigid Body items
    paramMiPlate = translateCaption("paramMiPlate");
    paramMiDisc = translateCaption("paramMiDisc");
    paramMiElementaryPlate = translateCaption("paramMiElementaryPlate");
    paramMiSphere = translateCaption("paramMiSphere");
    paramMiCuboid = translateCaption("paramMiCuboid");
    //menu Frame items
    paramMiBodyFixedFrame = translateCaption("paramMiBodyFixedFrame");
    paramMiLaboratoryFrame = translateCaption("paramMiLaboratoryFrame");
    //menu Projection items
    paramMiXYProjection = translateCaption("paramMiXYProjection");
    paramMiXZProjection = translateCaption("paramMiXZProjection");
    paramMiYZProjection = translateCaption("paramMiYZProjection");
    paramMi3DProjection = translateCaption("paramMi3DProjection");
    paramMi3DProjectionPerspective = translateCaption(
        "paramMi3DProjectionPerspective");
    //menu Show items
    paramMiBody = translateCaption("paramMiBody");
    paramMiBodyFixedAxes = translateCaption("paramMiBodyFixedAxes");
    paramMiLabAxes = translateCaption("paramMiLabAxes");
    paramMiOmegaL = translateCaption("paramMiOmegaL");
    paramMiTraces = translateCaption("paramMiTraces");
    //menu Help items
    paramMiHelp = translateCaption("paramMiHelp");
    paramMiAbout = translateCaption("paramMiAbout");
    //others
    paramPanel3DBodyTitle = translateCaption("paramPanel3DBodyTitle");
    paramPanel3DLabTitle = translateCaption("paramPanel3DLabTitle");
    paramSlidersTitle = translateCaption("paramSlidersTitle");
    paramNext = translateCaption("paramNext");
  }

  public String translateCaption(String parameterName) {
    return getParameter(parameterName);
  }

}