package org.opensourcephysics.tuleja.feynmanplate;

import org.opensourcephysics.numerics.*;
import org.opensourcephysics.displayejs.*;
import java.awt.Color;
import java.awt.Font;

/**
 * <p>Title: Sphere</p>
 * <p>Description: Subclass of RigidBody. It can draw itself and compute its motion.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * @author Alexander Tomori, Boris Gazovic, Slavomir Tuleja
 * @version 0.01
 */
public class Sphere extends RigidBody {
  double radius=0.5;
  int nPhi=20;//for plate
  int nTheta=20;
  double[][][] data=new double[nPhi][nTheta][3];
  double[][][] dataTransformed=new double[nPhi][nTheta][3];
  InteractiveSurface sphere;
  /**
   * Constructor of the sphere object.
   */
  public Sphere(){
    super(0.12567,0.12567);

    sphere=new InteractiveSurface();
    for(int i=0; i<nPhi; i++){
      for(int j=0; j<nTheta; j++){
        double theta=Math.PI* j/(nTheta-1.0);
        double phi=2*Math.PI* i/(nPhi-1.0);
        data[i][j][0]=radius* Math.sin(theta)* Math.cos(phi);
        data[i][j][1]=radius* Math.sin(theta)* Math.sin(phi);
        data[i][j][2]=radius* Math.cos(theta);
      }
    }
    sphere.getStyle().setEdgeColor(new Color(53, 97, 255));
    sphere.setColorOriginAndDirection(0.0, 0.0, 0.0, 0.0, 0.0, 5.0);
    sphere.setColorRegions(new double[]{}, new Color[]{new Color(198, 222, 255)});

    sphere.setData(data);
    sphere.setOrigin(0,0,0,false);

    add(sphere);
    setVisibility();
  }


  public void setVisibility(){
    super.setVisibilityExceptRigidBody();
    sphere.setVisible(showRigidBody);
  }

  public void transformBodyOnly(double[][] rm){
    for(int i=0; i<nPhi; i++){
      for(int j=0; j<nTheta; j++){
        double x=data[i][j][0];
        double y=data[i][j][1];
        double z=data[i][j][2];
        dataTransformed[i][j][0]=x*rm[0][0]+ y*rm[0][1]+ z*rm[0][2];
        dataTransformed[i][j][1]=x*rm[1][0]+ y*rm[1][1]+ z*rm[1][2];
        dataTransformed[i][j][2]=x*rm[2][0]+ y*rm[2][1]+ z*rm[2][2];
      }
    }
    sphere.setData(dataTransformed);
  }

}
