package org.opensourcephysics.tuleja.grorbits;

/** Object encapsulating differential equations for motion of LIGHT
 *  in equatorial plane of a Kerr black hole in Boyer-Lindquist bookkeeper's Coordinates
 *  For comments about the functionality of the methods below,
 *  please see the definition of GROrbitODE interface in GROrbitODE.java
 *  (c) Slavomir Tuleja, 2002
 */

public class BoyerLindquistLight implements GROrbitODE {
   private double[] state= new double[] {10.0,0.0,0.0,0.0,0.0};
   private double a=0;
   private double invB=0.5; // invB = 1/b
   private int sign=-1;//for inward orbit sign=-1; for outward orbit sign=+1
   private double rLocalMin, rLocalMax;
   private double vmMax=1, vmMin=-1;
   //...initial state: {r,dr/dt,phi,tau,t}
   private GRorbits applet;

   //constructor
   public BoyerLindquistLight(GRorbits applet){
      this.applet=applet;
   }

   public double[] getState(){
      return state;
   }

   public void getRates(double[] state, double[] rates){
      double b=1/invB;
      double r1;
      rates[0]=state[1];
      r1=-((a*a - 2*state[0] + state[0]*state[0])*(2*a*a*a*(a-b)*(a-b)*(a-b) + a*(a-b)*(a-b)*(3*a*a*a + 4*a - 4*b)*state[0] + a*(a - b)*(a*a*a*a + a*a*a*b - 4*a*a + 14*a*b - 10*b*b)*state[0]*state[0] - 2*a*b*(3*a*a - 5*a*b + 2*b*b)*state[0]*state[0]*state[0] + 2*(a*a*a*a + 6*a*b - 5*b*b - a*a*(b*b +1))* state[0]*state[0]*state[0]*state[0] + ((-5)*a*a  + 7*b*b)*state[0]*state[0]*state[0]*state[0]*state[0] + (a*a - b*b)*state[0]*state[0]*state[0]*state[0]*state[0]*state[0] - 2*state[0]*state[0]*state[0]*state[0]*state[0]*state[0]*state[0]))/( state[0]*state[0]*Math.pow(2*a*(a - b) + a*a*state[0] + state[0]*state[0]*state[0],3) );
      /*
      *this remains an unsolved problem...
      if(r1>1e9)r1=1e9;
      if(r1<-1e9)r1=-1e9;
      */
      rates[1]=r1;
      rates[2]=(2*a-2*b+b*state[0])/(2*a*a-2*a*b+a*a*state[0]+state[0]*state[0]*state[0]);
      rates[3]=0;
      rates[4]=1;
   }

   public void setA(double a){
      this.a=a;
   }

   public double getA(){
      return a;
   }

   public void setIC(double r, double phi){
      //sets and computes initial conditions r, dr/dt, phi, dphi/dt from
      //given vlaue of invB
      double b=1/invB;

      //double eps=1e-10;
      //if(Math.abs(r-getRHorizon())<eps)r=getRHorizon()+eps;

      state[0]=r;
      state[1]=sign*Math.sqrt((Math.pow(a*a+state[0]*state[0]-2*state[0],2)*(state[0]*state[0]*state[0]-4*a*b+b*b*(2-state[0])+a*a*(2+state[0])))/(state[0]*Math.pow(state[0]*state[0]*state[0]-2*a*b+a*a*(2+state[0]),2)));
      state[2]=phi;
      state[3]=0;
      state[4]=0;
   }

   public double getVm(double r){
      return (2*a+Math.sqrt(r*r*(a*a+r*r-2*r)))/(r*r*r+a*a*(r+2));
   }

   public double getVmAtHorizon(){
      return a/(2*(1+Math.sqrt(1-a*a)));
   }

   public double getRHorizon(){
      return 1+Math.sqrt(1-a*a);
   }

   public double getRInnerHorizon(){
      return 1-Math.sqrt(1-a*a);
   }

   public double getR(){
      return state[0];
   }

   public double getPhi(){
      return state[2];
   }

   public double getTau(){
      return state[3];
   }

   public Scale getScale(double rMax){
      //returns the scale object for Vm diagram
      //also it sets value of public variables rLocalMin, rLocalMax
      // if rLocalMin==0 then there's no local minimum in Vm(r)
      // else there's local minimum in Vm(r) at r=rLocalMin...
      // the same is true about rLocalMax
      double y1=0;
      double y2=1;
      double deltaY;
      double min,max,yA,yB,yC,rA,rB,rC;
      if(rMax<getRHorizon()){
         min=0;
         max=0.2;
      }
      else{
         double dr=rMax/400.0;

         min=1e14; max=-1e14;
         rLocalMin=0;
         rLocalMax=0;
         rA=getRHorizon()+1e-10; rB=rA+dr; rC=rA+2*dr;

         while (rA<=rMax-2*dr){
            double VmrA=getVm(rA);
            double VmrB=getVm(rB);
            double VmrC=getVm(rC);
            if(VmrA<min) {min=VmrA;}
            if(VmrA>max) {max=VmrA;}
            if((VmrB-VmrA)*(VmrC-VmrB)<0) {
               if((VmrB-VmrA)>0) rLocalMax=rB;
               else rLocalMin=rB;
            }
            rA+=dr;
            rB+=dr;
            rC+=dr;
         }
      }

      if(rLocalMin>getRHorizon()){
         //local minimum in Vm(r) exists...
         y1=getVm(rLocalMin);
         y2=max;
      }
      else {
         //there is no local minimum in Vm(r)...
         y1=min;
         y2=max;
      }

      setEffPotYBounds();

      if(y2>vmMax) y2=vmMax;
      if(y1<vmMin) y1=vmMin;

      deltaY=y2-y1;
      y1-=0.05*deltaY;
      y2+=0.05*deltaY;
      return new Scale(0,rMax,y1,y2,0,200,0,200);
   }

   private void setEffPotYBounds(){
      //sets vmMax and vmMin depending on applet zoom
      int zoom=applet.getZoom();
      switch(zoom){
         case 0: //x0.5
            vmMax=1;
            vmMin=0;
            return;
         case 1: //x1
            vmMax=1;
            vmMin=0;
            return;
         case 2: //x10
            vmMax=1;
            vmMin=0;
            return;
         case 3: //x100
            vmMax=1;
            vmMin=0;
            return;
         default:
            vmMax=1;
         vmMin=0;
         return;
      }
   }

   public void setSign(int s){
      sign=s;
   }

   public double getRLocalMin(double rMax){
      Scale sc=getScale(rMax);
      if(rLocalMin<getRHorizon()) return 0;
      else return findStationaryPosition(rLocalMin-rMax/400.0,rLocalMin+rMax/400);
   }

   public double getRLocalMax(double rMax){
      Scale sc=getScale(rMax);
      if(rLocalMax<getRHorizon()) return 0;
      else return findStationaryPosition(rLocalMax-rMax/400.0,rLocalMax+rMax/400);
   }

   private double findStationaryPosition(double a, double b){
      //we suppose that a < b
      double eps=1e-10;
      double newA=a, newB=b;
      double v1, v2;
      double minAbsSlope=1e12;
      double absSlope;

      double x=a;
      double dx=(b-a)/10.0;

      while(x < b) {
         v1=getVm(x);
         v2=getVm(x+dx);
         absSlope=Math.abs((v2-v1)/dx);
         if(absSlope<minAbsSlope){
            minAbsSlope=absSlope;
            newA=x;
            newB=x+dx;
         }

         x+=dx;
      }

      if(dx<eps) return (newA+newB)/2.0;
      else return findStationaryPosition(newA,newB);
   }

   public void setLm(double Lm){
      //does nothing...
   }

   public void setEm(double Em){
      //does nothing...
   }

   public double getOrbitParameter(){
      //returns Em or 1/b depending on ode selected...
      return invB;
   }

   public void setOrbitParameter(int j){
      invB=applet.scVm.jToy(j);
   }

   public double getLm(){
      //this is NOT useful here...
      return 0;
   }

   public void setInvB(double invB){
      this.invB=invB;
   }

   public double getEm(){
      //this is NOT useful here...
      return 0;
   }

   public double rTorPlot(double r){
      return Math.sqrt(r*r+a*a);
   }

   public double rPlotTor(double rPlot){
      if(rPlot*rPlot>a*a) return Math.sqrt(rPlot*rPlot-a*a);
      else return 0;
   }

   public void initialSetup(){
      //scrollbars and corresponding labels...
      a=applet.slA.getValue()/1000000.0;
      initialSetupTextual();
   }

   public void initialSetupTextual(){
      //scrollbars and corresponding labels...
      applet.lblA.setText("J/M = a = ".concat(MyUtils.roundOff(a,6)).concat(" M"));

      //inicialization of scale objects...
      applet.setScales(applet.getRMax());

      //adjusting invB...
      //...if invB is under Effective potential curve, we have to adjust it
      if(invB<getVm(applet.getR())) invB=getVm(applet.getR())+1e-10;

      //setting initial conditions of simulation
      setIC(applet.getR(),applet.getPhi());

      //array of orbit points
      for(int i=0; i<applet.numPoints;i++){
         applet.orbit[i]= new OrbitPoint(getR(),getPhi());
      }

      applet.repaintAll();
   }

}
