/*
 * put your module comment here
 * formatted with JxBeauty (c) johann.langhofer@nextra.at
 */


package org.opensourcephysics.tuleja.gyroscope;

/**
 * <p>Title: GyroscopeApp</p>
 * <p>Description: 3D simulation of a gyroscope suspended in double gimbals</p>
 * <p>Copyright: Copyright (c) August 1, 2003</p>
 * <p>Company: </p>
 * @author Slavomir Tuleja
 * @version 1.0
 */
import org.opensourcephysics.display3d.*;
import org.opensourcephysics.controls.*;
import org.opensourcephysics.numerics.*;
import org.opensourcephysics.display.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * put your documentation comment here
 */
public class GyroscopeApp
    implements ActionListener, Animation {
  DrawingPanel3D panel3D;
  DrawingFrame3D frame;
  DrawingPanel3D panel3DSphere;
  DrawingFrame3D frameSphere;
  DrawingFrame framePlot;
  PlottingPanel panelPlot;
  Dataset dataset;
  InteractiveCircle dot;
  Dataset3D dataset3D;
  Timer timer = new Timer(100, this);
  Control myControl;
  double r = 0.08, h = 0.04; //radius and height of a gyroscope cylinder
  double length; //distance of c.m. of the body cylinder from the fixed point (center of gimbals)
  double La = 0.24, Lb = 0.3; //dimensions of inner and outer gimbal
  double rAxle = 0.002; //radius of any axle
  double wGimbal = 0.005; //width of any gimbal bar
  double wFrame = 0.075; //width of supporting static frame
  double outerAxleLength = 0.06;
  double densBody = 7600; //density of material forming gimbals and the rotating body
  double m; //mass of the gyro cylinder
  double a, b, energy; //variables connected with Goldstein analysis
  double I, I2; //principal moments of inertia of the rotating body
  double g = 9.8;
  Color bodyColor = Color.red, axleColor = Color.yellow,
      gimbalColor = new Color(0.4f,
                              0.4f, 0.5f);
  DShapeGroup body, innerGimbal, outerGimbal, staticFrame;
  Gyroscope ode = new Gyroscope();
  ODEAdaptiveSolver odeSolver = new RK45MultiStep(ode);

  /**
   * put your documentation comment here
   */
  public GyroscopeApp() {
    panel3D = new DrawingPanel3D();
    frame = new DrawingFrame3D(panel3D);
    panel3DSphere = new DrawingPanel3D();
    frameSphere = new DrawingFrame3D(panel3DSphere);
    frameSphere.setSize(new Dimension(300, 300));
  }

  /**
   * put your documentation comment here
   */
  public void createSphereScene() {
    panel3DSphere.clear();
    dataset3D = new Dataset3D();
    dataset3D.setConnected(true);
    dataset3D.setMarkerShape(Dataset3D.NO_MARKER);
    DSphere sphere = new DSphere(0, 0, 0, 1, Color.white);
    sphere.setTransparency(0.3f);
    //DAxisSet axisSet = new DAxisSet(-1.5,1.5,0.5);
    panel3DSphere.addDrawable3D(sphere);
    panel3DSphere.addDrawable3D(dataset3D);
    panel3DSphere.setScenePEuler(1.1554, 0.1819, -1.6338);
    panel3DSphere.repaint();
    org.opensourcephysics.display.GUIUtils.showDrawingAndTableFrames();
  }

  /**
   * put your documentation comment here
   */
  public void createGyro3DScene() {
    panel3D.clear();
    DCylinder cylinder = new DCylinder(0, 0, 0, r, h, bodyColor);
    DCylinder bodyAxle = new DCylinder(0, 0, 0, rAxle, La - wGimbal,
                                       Color.yellow);
    DCylinder mark = new DCylinder(r, 0, 0.0, 0.001, h, Color.white);
    body = new DShapeGroup();
    //body.addShape(bodyAxle);
    body.addShape(cylinder);
    body.addShape(mark);
    //create the innerGimbal DShapeGroup
    DBox innerBar1 = new DBox(0, La / 2., 0, La + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar2 = new DBox(0, -La / 2., 0, La + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar3 = new DBox( -La / 2., 0, 0, wGimbal, La - wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar4 = new DBox(La / 2., 0, 0, wGimbal, La - wGimbal, wGimbal,
                              gimbalColor);
    DCylinder innerAxle1 = new DCylinder( (La / 2. + Lb / 2.) / 2., 0, 0, rAxle,
                                         Lb / 2. - La / 2. - wGimbal, axleColor);
    innerAxle1.rotate(0, 0, Math.PI / 2.);
    DCylinder innerAxle2 = new DCylinder( - (La / 2. + Lb / 2.) / 2., 0, 0,
                                         rAxle,
                                         Lb / 2. - La / 2. - wGimbal, axleColor);
    innerAxle2.rotate(0, 0, Math.PI / 2.);
    innerGimbal = new DShapeGroup();
    innerGimbal.addShape(innerBar1);
    innerGimbal.addShape(innerBar2);
    innerGimbal.addShape(innerBar3);
    innerGimbal.addShape(innerBar4);
    innerGimbal.addShape(innerAxle1);
    innerGimbal.addShape(innerAxle2);
    innerGimbal.addShape(bodyAxle);
    //create the outerGimbal DShapeGroup
    DBox outerBar1 = new DBox(0, Lb / 2., 0, Lb + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar2 = new DBox(0, -Lb / 2., 0, Lb + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar3 = new DBox( -Lb / 2., 0, 0, wGimbal, Lb - wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar4 = new DBox(Lb / 2., 0, 0, wGimbal, Lb - wGimbal, wGimbal,
                              gimbalColor);
    DCylinder outerAxle1 = new DCylinder(0,
                                         Lb / 2. + wGimbal / 2. +
                                         outerAxleLength / 2.,
                                         0, rAxle, outerAxleLength, axleColor);
    DCylinder outerAxle2 = new DCylinder(0,
                                         -Lb / 2. - wGimbal / 2. -
                                         outerAxleLength / 2.,
                                         0, rAxle, outerAxleLength, axleColor);
    outerGimbal = new DShapeGroup();
    outerGimbal.addShape(outerBar1);
    outerGimbal.addShape(outerBar2);
    outerGimbal.addShape(outerBar3);
    outerGimbal.addShape(outerBar4);
    outerGimbal.addShape(outerAxle1);
    outerGimbal.addShape(outerAxle2);
    //create the staticFrame DShapeGroup
    DBox top = new DBox(0, Lb / 2. + outerAxleLength + wGimbal, 0, wFrame,
                        2 * wGimbal, wFrame, gimbalColor);
    top.setTransparency(0.5f);
    DBox base = new DBox(0, -Lb / 2. - outerAxleLength - wGimbal, 0, wFrame,
                         2 * wGimbal, wFrame, gimbalColor);
    base.setTransparency(0.5f);
    staticFrame = new DShapeGroup();
    staticFrame.addShape(top);
    staticFrame.addShape(base);
    //add everything to panel3D
    panel3D.addDrawable3D(body);
    panel3D.addDrawable3D(innerGimbal);
    panel3D.addDrawable3D(outerGimbal);
    panel3D.addDrawable3D(staticFrame);
    panel3D.setSceneXYZ(0, 0, -0.94656);
  }

  /**
   * put your documentation comment here
   * @param evt
   */
  public void actionPerformed(ActionEvent evt) {
    stepAnimation();
  }

  /**
   * put your documentation comment here
   */
  public void initializeAnimation() {

    length =  myControl.getDouble("offset");
    I = Math.PI * r * r * h * densBody / 12. *
        (h * h + 3 * r * r + 12 * length * length);
    I2 = 1. / 2. * Math.PI * densBody * h * r * r * r * r;
    m = Math.PI * r * r * h * densBody;
    ode.setI(I, I2);
    ode.setM(m);
    ode.setL(length);

    double theta = Math.toRadians(myControl.getDouble("theta"));
    double phi = Math.toRadians(myControl.getDouble("phi"));
    double psi = Math.toRadians(myControl.getDouble("psi"));
    double thetaD = Math.toRadians(myControl.getDouble("thetaDot"));
    double phiD = Math.toRadians(myControl.getDouble("phiDot"));
    double psiD = Math.toRadians(myControl.getDouble("psiDot"));

    ode.setIC(theta, phi, psi, thetaD, phiD, psiD, 0);
    odeSolver.initialize(0.01);
    odeSolver.setTolerance(1e-6);

    createGyro3DScene();
    createSphereScene();
    body.setPEuler(phi + Math.PI / 2., theta, psi - Math.PI / 2.);
    body.setXYZ(length * Math.sin(theta) * Math.sin(phi),
                length * Math.cos(theta),
                length * Math.sin(theta) * Math.cos(phi));
    innerGimbal.setPEuler(phi + Math.PI / 2., theta, -Math.PI / 2.);
    outerGimbal.setPEuler(phi + Math.PI / 2., 0, -Math.PI / 2.);
  }

  /**
   * put your documentation comment here
   */
  public void resetAnimation() {
    myControl.clearValues();
    myControl.setValue("offset", "0.05");
    myControl.setValue("theta", "30");
    myControl.setValue("phi", "30");
    myControl.setValue("psi", "250");
    myControl.setValue("thetaDot", "30");
    myControl.setValue("phiDot", "0");
    myControl.setValue("psiDot", "1750");
  }

  /**
   * put your documentation comment here
   */
  public void stepAnimation() {
    double phi, psi, theta;
    odeSolver.step();
    phi = ode.getPhi();
    psi = ode.getPsi();
    theta = ode.getTheta();
    body.setPEuler(phi + Math.PI / 2., theta, psi - Math.PI / 2.);
    body.setXYZ(length * Math.sin(theta) * Math.sin(phi),
                length * Math.cos(theta),
                length * Math.sin(theta) * Math.cos(phi));
    innerGimbal.setPEuler(phi + Math.PI / 2., theta, -Math.PI / 2.);
    outerGimbal.setPEuler(phi, 0, 0);
    panel3D.repaint();
    dataset3D.append(Math.sin(phi) * Math.sin(theta), Math.cos(theta),
                     Math.cos(phi) * Math.sin(theta));
    panel3DSphere.repaint();
  }

  /**
   * put your documentation comment here
   */
  public void startAnimation() {
    timer.start();
  }

  /**
   * put your documentation comment here
   */
  public void stopAnimation() {
    timer.stop();
  }

  /**
   * put your documentation comment here
   * @param control
   */
  public void setControl(Control control) {
    myControl = control;
    resetAnimation();
  }



  /**
   * put your documentation comment here
   * @param args
   */
  public static void main(String[] args) {
    Animation model = new GyroscopeApp();
    Control control = new AnimationControl(model);
    model.setControl(control);
  }
}
