package org.opensourcephysics.tuleja.gyroscope;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import org.opensourcephysics.display3d.*;
import org.opensourcephysics.display3d.*;
import org.opensourcephysics.numerics.*;

public class JGyroscope
    extends JApplet
    implements ActionListener, ChangeListener {

  Container cp;
  JPanel pnlSliders, pnlSettings, pnlEast;
  JButton btnStartStop;
  JSlider slLength, slTheta, slPhi, slPsi, slThetaD, slPhiD, slPsiD;
  DrawingPanel3D panel3D;
  javax.swing.Timer timer;
  ActionListener tickingHandler;

  double r = 0.08, h = 0.04; //radius and height of a gyroscope cylinder
  double length; //distance of c.m. of the body cylinder from the fixed point (center of gimbals)
  double La = 0.24, Lb = 0.3; //dimensions of inner and outer gimbal
  double rAxle = 0.002; //radius of any axle
  double wGimbal = 0.005; //width of any gimbal bar
  double wFrame = 0.075; //width of supporting static frame
  double outerAxleLength = 0.06;
  double densBody = 7600; //density of material forming gimbals and the rotating body
  double m; //mass of the gyro cylinder
  double I, I2; //principal moments of inertia of the rotating body
  double g = 9.8;
  Color bodyColor = Color.red, axleColor = Color.yellow,
      gimbalColor = new Color(0.4f,
                              0.4f, 0.5f);
  DShapeGroup body, innerGimbal, outerGimbal, staticFrame;
  Gyroscope ode = new Gyroscope();
  ODEAdaptiveSolver odeSolver = new RK45MultiStep(ode);

  /**
   *
   */
  public void init() {
    //Creating GUI
    panel3D = new DrawingPanel3D();
    panel3D.setBorder(BorderFactory.createTitledBorder("3D Universe"));

    //button
    btnStartStop = new JButton("Start");
    btnStartStop.addActionListener(this);

    //sliders
    slLength = new JSlider(JSlider.HORIZONTAL, -70, 70, 70);
    slLength.setBorder(BorderFactory.createTitledBorder("offset = " +
        slLength.getValue() + " milimeters"));
    slLength.addChangeListener(this);

    slTheta = new JSlider(JSlider.HORIZONTAL, 1, 179, 30);
    slTheta.setBorder(BorderFactory.createTitledBorder("\u03B8 = " +
        slTheta.getValue() + " "));
    slTheta.addChangeListener(this);

    slPhi = new JSlider(JSlider.HORIZONTAL, 0, 360, 30);
    slPhi.setBorder(BorderFactory.createTitledBorder("\u03D5 = " +
        slPhi.getValue() + " "));
    slPhi.addChangeListener(this);

    slPsi = new JSlider(JSlider.HORIZONTAL, 0, 360, 250);
    slPsi.setBorder(BorderFactory.createTitledBorder("\u03C8 = " +
        slPsi.getValue() + " "));
    slPsi.addChangeListener(this);

    slThetaD = new JSlider(JSlider.HORIZONTAL, -500, 500, 30);
    slThetaD.setBorder(BorderFactory.createTitledBorder("d\u03B8/dt = " +
        slThetaD.getValue() + " /s"));
    slThetaD.addChangeListener(this);

    slPhiD = new JSlider(JSlider.HORIZONTAL, -500, 500, 0);
    slPhiD.setBorder(BorderFactory.createTitledBorder("d\u03D5/dt = " +
        slPhiD.getValue() + " /s"));
    slPhiD.addChangeListener(this);

    slPsiD = new JSlider(JSlider.HORIZONTAL, -3600, 3600, 1750);
    slPsiD.setBorder(BorderFactory.createTitledBorder("d\u03C8/dt = " +
        slPsiD.getValue() + " /s"));

    slPsiD.addChangeListener(this);



    pnlSliders = new JPanel();
    pnlSliders.setBorder(BorderFactory.createTitledBorder("Initial conditions:"));
    pnlSliders.setLayout(new GridLayout(6, 1));
    pnlSliders.add(slTheta);
    pnlSliders.add(slPhi);
    pnlSliders.add(slPsi);
    pnlSliders.add(slThetaD);
    pnlSliders.add(slPhiD);
    pnlSliders.add(slPsiD);


    pnlSettings = new JPanel();
    pnlSettings.setBorder(BorderFactory.createTitledBorder("Settings:"));
    pnlSettings.setLayout(new GridLayout(1, 1));
    pnlSettings.add(slLength);

    pnlEast = new JPanel();
    pnlEast.setLayout(new BoxLayout(pnlEast,BoxLayout.Y_AXIS));
    pnlEast.add(pnlSliders);
    pnlEast.add(pnlSettings);

    cp = getContentPane();

    cp.setLayout(new BorderLayout());
    cp.add(panel3D, BorderLayout.CENTER);
    cp.add(pnlEast, BorderLayout.EAST);
    cp.add(btnStartStop, BorderLayout.SOUTH);
    cp.validate();
    cp.repaint();
    initializeAnimation();

    tickingHandler = new TickingHandler(this);
    timer = new javax.swing.Timer(100, tickingHandler);
  }


  public static void main(String[] args) {
    JFrame fr = new JFrame("JGyroscope");
    fr.setSize(700, 500);
    JGyroscope applet = new JGyroscope();
    applet.init();
    applet.start();
    fr.getContentPane().add(applet, BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }


  /**
   *
   */
  public void start() {}

  /**
   * This is done each time the Timer ticks
   * @param evt ActionEvent
   */
  public void actionPerformed(ActionEvent evt) {
    if (evt.getActionCommand().equals("Start")) {
      btnStartStop.setText("Stop");
      slTheta.setEnabled(false);
      slPhi.setEnabled(false);
      slPsi.setEnabled(false);
      slThetaD.setEnabled(false);
      slPhiD.setEnabled(false);
      slPsiD.setEnabled(false);
      slLength.setEnabled(false);
      startAnimation();
    }
    else if (evt.getActionCommand().equals("Stop")) {
      btnStartStop.setText("Start");
      stopAnimation();
      slTheta.setEnabled(true);
      slPhi.setEnabled(true);
      slPsi.setEnabled(true);
      slThetaD.setEnabled(true);
      slPhiD.setEnabled(true);
      slPsiD.setEnabled(true);
      slLength.setEnabled(true);

    }
  }

  /**
   *
   */
  public void createGyro3DScene() {
    panel3D.clear();
    DCylinder cylinder = new DCylinder(0, 0, 0, r, h, bodyColor);
    DCylinder bodyAxle = new DCylinder(0, 0, 0, rAxle, La - wGimbal,
                                       Color.yellow);
    DCylinder mark = new DCylinder(r, 0, 0.0, 0.001, h, Color.white);
    body = new DShapeGroup();
    //body.addShape(bodyAxle);
    body.addShape(cylinder);
    body.addShape(mark);
    //create the innerGimbal DShapeGroup
    DBox innerBar1 = new DBox(0, La / 2., 0, La + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar2 = new DBox(0, -La / 2., 0, La + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar3 = new DBox( -La / 2., 0, 0, wGimbal, La - wGimbal, wGimbal,
                              gimbalColor);
    DBox innerBar4 = new DBox(La / 2., 0, 0, wGimbal, La - wGimbal, wGimbal,
                              gimbalColor);
    DCylinder innerAxle1 = new DCylinder( (La / 2. + Lb / 2.) / 2., 0, 0, rAxle,
                                         Lb / 2. - La / 2. - wGimbal, axleColor);
    innerAxle1.rotate(0, 0, Math.PI / 2.);
    DCylinder innerAxle2 = new DCylinder( - (La / 2. + Lb / 2.) / 2., 0, 0,
                                         rAxle,
                                         Lb / 2. - La / 2. - wGimbal, axleColor);
    innerAxle2.rotate(0, 0, Math.PI / 2.);
    innerGimbal = new DShapeGroup();
    innerGimbal.addShape(innerBar1);
    innerGimbal.addShape(innerBar2);
    innerGimbal.addShape(innerBar3);
    innerGimbal.addShape(innerBar4);
    innerGimbal.addShape(innerAxle1);
    innerGimbal.addShape(innerAxle2);
    innerGimbal.addShape(bodyAxle);
    //create the outerGimbal DShapeGroup
    DBox outerBar1 = new DBox(0, Lb / 2., 0, Lb + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar2 = new DBox(0, -Lb / 2., 0, Lb + wGimbal, wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar3 = new DBox( -Lb / 2., 0, 0, wGimbal, Lb - wGimbal, wGimbal,
                              gimbalColor);
    DBox outerBar4 = new DBox(Lb / 2., 0, 0, wGimbal, Lb - wGimbal, wGimbal,
                              gimbalColor);
    DCylinder outerAxle1 = new DCylinder(0,
                                         Lb / 2. + wGimbal / 2. +
                                         outerAxleLength / 2.,
                                         0, rAxle, outerAxleLength, axleColor);
    DCylinder outerAxle2 = new DCylinder(0,
                                         -Lb / 2. - wGimbal / 2. -
                                         outerAxleLength / 2.,
                                         0, rAxle, outerAxleLength, axleColor);
    outerGimbal = new DShapeGroup();
    outerGimbal.addShape(outerBar1);
    outerGimbal.addShape(outerBar2);
    outerGimbal.addShape(outerBar3);
    outerGimbal.addShape(outerBar4);
    outerGimbal.addShape(outerAxle1);
    outerGimbal.addShape(outerAxle2);
    //create the staticFrame DShapeGroup
    DBox top = new DBox(0, Lb / 2. + outerAxleLength + wGimbal, 0, wFrame,
                        2 * wGimbal, wFrame, gimbalColor);
    top.setTransparency(0.5f);
    DBox base = new DBox(0, -Lb / 2. - outerAxleLength - wGimbal, 0, wFrame,
                         2 * wGimbal, wFrame, gimbalColor);
    base.setTransparency(0.5f);
    staticFrame = new DShapeGroup();
    staticFrame.addShape(top);
    staticFrame.addShape(base);
    //add everything to panel3D
    panel3D.addDrawable3D(body);
    panel3D.addDrawable3D(innerGimbal);
    panel3D.addDrawable3D(outerGimbal);
    panel3D.addDrawable3D(staticFrame);
    panel3D.setSceneXYZ(0, 0, -0.94656);
  }

  /**
   *
   */
  public void initializeAnimation() {
    length = ( (double) slLength.getValue()) / 1000.0;
    I = Math.PI * r * r * h * densBody / 12. *
        (h * h + 3 * r * r + 12 * length * length);
    I2 = 1. / 2. * Math.PI * densBody * h * r * r * r * r;
    m = Math.PI * r * r * h * densBody;
    ode.setI(I, I2);
    ode.setM(m);
    ode.setL(length);

    double theta = Math.toRadians(slTheta.getValue());
    double phi = Math.toRadians(slPhi.getValue());
    double psi = Math.toRadians(slPsi.getValue());
    double thetaD = Math.toRadians(slThetaD.getValue());
    double phiD = Math.toRadians(slPhiD.getValue());
    double psiD = Math.toRadians(slPsiD.getValue());

    ode.setIC(theta, phi, psi, thetaD, phiD, psiD, 0);
    odeSolver.initialize(0.01);
    odeSolver.setTolerance(1e-6);

    createGyro3DScene();
    body.setPEuler(phi + Math.PI / 2., theta, psi - Math.PI / 2.);
    body.setXYZ(length * Math.sin(theta) * Math.sin(phi),
                length * Math.cos(theta),
                length * Math.sin(theta) * Math.cos(phi));
    innerGimbal.setPEuler(phi + Math.PI / 2., theta, -Math.PI / 2.);
    outerGimbal.setPEuler(phi + Math.PI / 2., 0, -Math.PI / 2.);

  }

  public void updateScene() {
    length = ( (double) slLength.getValue()) / 1000.0;
    I = Math.PI * r * r * h * densBody / 12. *
        (h * h + 3 * r * r + 12 * length * length);
    I2 = 1. / 2. * Math.PI * densBody * h * r * r * r * r;
    m = Math.PI * r * r * h * densBody;
    ode.setI(I, I2);
    ode.setM(m);
    ode.setL(length);

    double theta = Math.toRadians(slTheta.getValue());
    double phi = Math.toRadians(slPhi.getValue());
    double psi = Math.toRadians(slPsi.getValue());
    double thetaD = Math.toRadians(slThetaD.getValue());
    double phiD = Math.toRadians(slPhiD.getValue());
    double psiD = Math.toRadians(slPsiD.getValue());

    ode.setIC(theta, phi, psi, thetaD, phiD, psiD, 0);
    odeSolver.initialize(0.01);
    odeSolver.setTolerance(1e-6);

    body.setPEuler(phi + Math.PI / 2., theta, psi - Math.PI / 2.);
    body.setXYZ(length * Math.sin(theta) * Math.sin(phi),
                length * Math.cos(theta),
                length * Math.sin(theta) * Math.cos(phi));
    innerGimbal.setPEuler(phi + Math.PI / 2., theta, -Math.PI / 2.);
    outerGimbal.setPEuler(phi + Math.PI / 2., 0, -Math.PI / 2.);
  }

  /**
   *
   */
  public void stepAnimation() {
    double phi, psi, theta;
    odeSolver.step();
    phi = ode.getPhi();
    psi = ode.getPsi();
    theta = ode.getTheta();
    body.setPEuler(phi + Math.PI / 2., theta, psi - Math.PI / 2.);
    body.setXYZ(length * Math.sin(theta) * Math.sin(phi),
                length * Math.cos(theta),
                length * Math.sin(theta) * Math.cos(phi));
    innerGimbal.setPEuler(phi + Math.PI / 2., theta, -Math.PI / 2.);
    outerGimbal.setPEuler(phi, 0, 0);
    panel3D.repaint();
  }

  /**
   *
   */
  public void startAnimation() {
    timer.start();
  }

  /**
   *
   */
  public void stopAnimation() {
    timer.stop();
  }

  public void stateChanged(ChangeEvent e) {
    JSlider source = (JSlider) e.getSource();

    if (source.equals(slLength)) {
      slLength.setBorder(BorderFactory.createTitledBorder("offset = " +
          slLength.getValue() + " milimeters"));
      updateScene();
    }
    else if (source.equals(slTheta)) {
      slTheta.setBorder(BorderFactory.createTitledBorder("\u03B8 = " +
          slTheta.getValue() + " "));
      updateScene();
    }
    else if (source.equals(slPhi)) {
      slPhi.setBorder(BorderFactory.createTitledBorder("\u03D5 = " +
          slPhi.getValue() + " "));
      updateScene();
    }
    else if (source.equals(slPsi)) {
      slPsi.setBorder(BorderFactory.createTitledBorder("\u03C8 = " +
          slPsi.getValue() + " "));
      updateScene();
    }
    else if (source.equals(slThetaD)) {
      slThetaD.setBorder(BorderFactory.createTitledBorder("d\u03B8/dt = " +
          slThetaD.getValue() + " /s"));
      updateScene();
    }
    else if (source.equals(slPhiD)) {
      slPhiD.setBorder(BorderFactory.createTitledBorder("d\u03D5/dt = " +
          slPhiD.getValue() + " /s"));
      updateScene();
    }
    else if (source.equals(slPsiD)) {
      slPsiD.setBorder(BorderFactory.createTitledBorder("d\u03C8/dt = " +
          slPsiD.getValue() + " /s"));
      updateScene();
    }

  }

}
