package org.opensourcephysics.tuleja.hypercube;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import org.opensourcephysics.display3d.*;
import javax.vecmath.*;

public class JHypercube
    extends JApplet
    implements ChangeListener {

  Container cp;
  JPanel pnlSouth, pnlSliders, pnlButtons, pnlHelpButtons;
  JSlider slAlpha12, slAlpha13, slAlpha14, slAlpha23, slAlpha24, slAlpha34;
  double[][] rot12, rot13, rot14, rot23, rot24, rot34, rot; //six 4-dimensional rotation matrices and one matrix rot equal to their composition
  double alpha12, alpha13, alpha14, alpha23, alpha24, alpha34; //six different angles of rotation in six different planes
  double d = 2; //x4-coordinate of center of projection C in 4D hyperspace -- C=[0,0,0,d]
  DrawingPanel3D panel = new DrawingPanel3D();
  Edge[] edge1, edge2, edge3, edge4; //arrays of edges with four different directions
  DVectorLine[] e1 = new DVectorLine[8]; //vectors representing projections of edges
  DVectorLine[] e2 = new DVectorLine[8];
  DVectorLine[] e3 = new DVectorLine[8];
  DVectorLine[] e4 = new DVectorLine[8];
  double w = 0.02; //width of vectors


  public static void main(String[] args) {
    JFrame fr = new JFrame("JHypercube");
    fr.setSize(400, 500);
    JHypercube applet = new JHypercube();
    applet.init();
    applet.start();
    fr.getContentPane().add(applet, BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }


  /**
   * Constructor: Sets all angles of rotation equal to 0. Initializes the virtual scene,
   * creates corresponding rotation matrices, creates array of edges in 4D hyperspace,
   * adds all edges to the scene.
   */
  public void init() {
    //Creating GUI

    //sliders
    slAlpha12 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha12.setBorder(BorderFactory.createTitledBorder("alpha12 = " +
        slAlpha12.getValue() + ""));
    slAlpha12.addChangeListener(this);

    slAlpha13 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha13.setBorder(BorderFactory.createTitledBorder("alpha13 = " +
        slAlpha13.getValue() + ""));
    slAlpha13.addChangeListener(this);

    slAlpha14 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha14.setBorder(BorderFactory.createTitledBorder("alpha14 = " +
        slAlpha14.getValue() + ""));
    slAlpha14.addChangeListener(this);

    slAlpha23 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha23.setBorder(BorderFactory.createTitledBorder("alpha23 = " +
        slAlpha23.getValue() + ""));
    slAlpha23.addChangeListener(this);

    slAlpha24 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha24.setBorder(BorderFactory.createTitledBorder("alpha24 = " +
        slAlpha24.getValue() + ""));
    slAlpha24.addChangeListener(this);

    slAlpha34 = new JSlider(JSlider.HORIZONTAL, 0, 360, 0);
    slAlpha34.setBorder(BorderFactory.createTitledBorder("alpha34 = " +
        slAlpha34.getValue() + ""));
    slAlpha34.addChangeListener(this);

    panel = new DrawingPanel3D();
    panel.setBorder(BorderFactory.createTitledBorder("3D Universe"));

    pnlSliders = new JPanel();
    pnlSliders.setLayout(new GridLayout(2, 3));
    pnlSliders.add(slAlpha12);
    pnlSliders.add(slAlpha13);
    pnlSliders.add(slAlpha14);
    pnlSliders.add(slAlpha23);
    pnlSliders.add(slAlpha24);
    pnlSliders.add(slAlpha34);
    pnlSliders.setBorder(BorderFactory.createTitledBorder(
        "Angles of Hypercube rotation"));

    pnlSouth = new JPanel();
    pnlSouth.setLayout(new BoxLayout(pnlSouth, BoxLayout.Y_AXIS));
    pnlSouth.add(pnlSliders);

    cp = getContentPane();

    cp.setLayout(new BorderLayout());
    cp.add(pnlSouth, BorderLayout.SOUTH);
    cp.add(panel, BorderLayout.CENTER);
    cp.validate();
    cp.repaint();

    alpha12 = Math.toRadians(0);
    alpha13 = Math.toRadians(0);
    alpha14 = Math.toRadians(0);
    alpha23 = Math.toRadians(0);
    alpha24 = Math.toRadians(0);
    alpha34 = Math.toRadians(0);

    initialize3DProjection();
    createRotationMatrices();
    createEdges();
    transformEdges();
    update3DProjection();

    org.opensourcephysics.display.GUIUtils.showDrawingAndTableFrames();
  }

  public void start() {

  }

  public void stateChanged(ChangeEvent e) {
    JSlider source = (JSlider) e.getSource();

    if (source.equals(slAlpha12)) {
      slAlpha12.setBorder(BorderFactory.createTitledBorder("alpha12 = " +
          slAlpha12.getValue() + ""));
      alpha12 = Math.toRadians(slAlpha12.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }
    else if (source.equals(slAlpha13)) {
      slAlpha13.setBorder(BorderFactory.createTitledBorder("alpha13 = " +
          slAlpha13.getValue() + ""));
      alpha13 = Math.toRadians(slAlpha13.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }
    else if (source.equals(slAlpha14)) {
      slAlpha14.setBorder(BorderFactory.createTitledBorder("alpha14 = " +
          slAlpha14.getValue() + ""));
      alpha14 = Math.toRadians(slAlpha14.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }
    else if (source.equals(slAlpha23)) {
      slAlpha23.setBorder(BorderFactory.createTitledBorder("alpha23 = " +
          slAlpha23.getValue() + ""));
      alpha23 = Math.toRadians(slAlpha23.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }
    else if (source.equals(slAlpha24)) {
      slAlpha24.setBorder(BorderFactory.createTitledBorder("alpha24 = " +
          slAlpha24.getValue() + ""));
      alpha24 = Math.toRadians(slAlpha24.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }
    else if (source.equals(slAlpha34)) {
      slAlpha34.setBorder(BorderFactory.createTitledBorder("alpha34 = " +
          slAlpha34.getValue() + ""));
      alpha34 = Math.toRadians(slAlpha34.getValue());
      createRotationMatrices();
      createEdges();
      transformEdges();
      update3DProjection();
    }

  }

  /**
   * Rotation matrices for six different possible rotations in hyperspace are defined.
   * Also their product, matrix "rot" is computed. It corresponds to the composition of
   * all six subsequent rotations.
   */
  public void createRotationMatrices() {
    double a, b;

    rot12 = new double[4][4];
    rot13 = new double[4][4];
    rot14 = new double[4][4];
    rot23 = new double[4][4];
    rot24 = new double[4][4];
    rot34 = new double[4][4];
    rot = new double[4][4];

    for (int i = 0; i < 4; i++) {
      for (int j = 0; j < 4; j++) {
        rot12[i][j] = 0;
        rot13[i][j] = 0;
        rot14[i][j] = 0;
        rot23[i][j] = 0;
        rot24[i][j] = 0;
        rot34[i][j] = 0;
        rot[i][j] = 0;
      }
    }

    //rot12
    a = Math.cos(alpha12);
    b = Math.sin(alpha12);
    rot12[0][0] = a;
    rot12[1][1] = a;
    rot12[0][1] = -b;
    rot12[1][0] = b;
    rot12[2][2] = 1;
    rot12[3][3] = 1;

    //rot13
    a = Math.cos(alpha13);
    b = Math.sin(alpha13);
    rot13[0][0] = a;
    rot13[2][2] = a;
    rot13[0][2] = -b;
    rot13[2][0] = b;
    rot13[1][1] = 1;
    rot13[3][3] = 1;

    //rot14
    a = Math.cos(alpha14);
    b = Math.sin(alpha14);
    rot14[0][0] = a;
    rot14[3][3] = a;
    rot14[0][3] = -b;
    rot14[3][0] = b;
    rot14[2][2] = 1;
    rot14[1][1] = 1;

    //rot23
    a = Math.cos(alpha23);
    b = Math.sin(alpha23);
    rot23[1][1] = a;
    rot23[2][2] = a;
    rot23[1][2] = -b;
    rot23[2][1] = b;
    rot23[0][0] = 1;
    rot23[3][3] = 1;

    //rot24
    a = Math.cos(alpha24);
    b = Math.sin(alpha24);
    rot24[1][1] = a;
    rot24[3][3] = a;
    rot24[1][3] = -b;
    rot24[3][1] = b;
    rot24[2][2] = 1;
    rot24[0][0] = 1;

    //rot34
    a = Math.cos(alpha34);
    b = Math.sin(alpha34);
    rot34[2][2] = a;
    rot34[3][3] = a;
    rot34[2][3] = -b;
    rot34[3][2] = b;
    rot34[0][0] = 1;
    rot34[1][1] = 1;

    //rot = rot34*rot24*rot23*rot14*rot13*rot12
    rot = multiplyMatrices(rot34, rot24);
    rot = multiplyMatrices(rot, rot23);
    rot = multiplyMatrices(rot, rot14);
    rot = multiplyMatrices(rot, rot13);
    rot = multiplyMatrices(rot, rot12);

  }

  /**
   * Creates arrays of edges of the unrotated 4D-hypercube. Edges are grouped according
   * to their direction, i.e. we have edges pointing in x1 direction in array edge1[] and so on.
   * Each direction will later be represented by different color.
   * An edge is defined by 8 numbers. First 4 numbers are 4-coordinates of the first point of the
   * edge and following 4 numbers represent the terminating point.
   */
  public void createEdges() {
    double a = 0.5;
    edge1 = new Edge[8];
    edge2 = new Edge[8];
    edge3 = new Edge[8];
    edge4 = new Edge[8];

    //8 edges in direction x1
    edge1[0] = new Edge( -a, -a, -a, -a, a, -a, -a, -a);
    edge1[1] = new Edge( -a, -a, -a, a, a, -a, -a, a);
    edge1[2] = new Edge( -a, -a, a, -a, a, -a, a, -a);
    edge1[3] = new Edge( -a, -a, a, a, a, -a, a, a);
    edge1[4] = new Edge( -a, a, -a, -a, a, a, -a, -a);
    edge1[5] = new Edge( -a, a, -a, a, a, a, -a, a);
    edge1[6] = new Edge( -a, a, a, -a, a, a, a, -a);
    edge1[7] = new Edge( -a, a, a, a, a, a, a, a);

    //8 edges in direction x2
    for (int i = 0; i < 8; i++) {
      edge2[i] = new Edge(edge1[i].a2, edge1[i].a1, edge1[i].a3, edge1[i].a4,
                          edge1[i].b2, edge1[i].b1, edge1[i].b3, edge1[i].b4);
    }

    //8 edges in direction x3
    for (int i = 0; i < 8; i++) {
      edge3[i] = new Edge(edge1[i].a3, edge1[i].a2, edge1[i].a1, edge1[i].a4,
                          edge1[i].b3, edge1[i].b2, edge1[i].b1, edge1[i].b4);
    }

    //8 edges in direction x4
    for (int i = 0; i < 8; i++) {
      edge4[i] = new Edge(edge1[i].a4, edge1[i].a2, edge1[i].a3, edge1[i].a1,
                          edge1[i].b4, edge1[i].b2, edge1[i].b3, edge1[i].b1);
    }
  }

  /**
   * This method takes all the edges of unrotated hypercube and using the rotation matrix
   * rot it creates edges of a rotated hypercube.
   */
  public void transformEdges() {
    Edge temp;
    double[] a = new double[4];
    double[] b = new double[4];

    //transform edges in x1 direction
    for (int n = 0; n < 8; n++) {
      for (int i = 0; i < 4; i++) {
        a[i] = rot[i][0] * edge1[n].a1 + rot[i][1] * edge1[n].a2 +
            rot[i][2] * edge1[n].a3 + rot[i][3] * edge1[n].a4;
        b[i] = rot[i][0] * edge1[n].b1 + rot[i][1] * edge1[n].b2 +
            rot[i][2] * edge1[n].b3 + rot[i][3] * edge1[n].b4;
      }
      edge1[n] = new Edge(a[0], a[1], a[2], a[3], b[0], b[1], b[2], b[3]);
    }

    //transform edges in x2 direction
    for (int n = 0; n < 8; n++) {
      for (int i = 0; i < 4; i++) {
        a[i] = rot[i][0] * edge2[n].a1 + rot[i][1] * edge2[n].a2 +
            rot[i][2] * edge2[n].a3 + rot[i][3] * edge2[n].a4;
        b[i] = rot[i][0] * edge2[n].b1 + rot[i][1] * edge2[n].b2 +
            rot[i][2] * edge2[n].b3 + rot[i][3] * edge2[n].b4;
      }
      edge2[n] = new Edge(a[0], a[1], a[2], a[3], b[0], b[1], b[2], b[3]);
    }

    //transform edges in x3 direction
    for (int n = 0; n < 8; n++) {
      for (int i = 0; i < 4; i++) {
        a[i] = rot[i][0] * edge3[n].a1 + rot[i][1] * edge3[n].a2 +
            rot[i][2] * edge3[n].a3 + rot[i][3] * edge3[n].a4;
        b[i] = rot[i][0] * edge3[n].b1 + rot[i][1] * edge3[n].b2 +
            rot[i][2] * edge3[n].b3 + rot[i][3] * edge3[n].b4;
      }
      edge3[n] = new Edge(a[0], a[1], a[2], a[3], b[0], b[1], b[2], b[3]);
    }

    //transform edges in x4 direction
    for (int n = 0; n < 8; n++) {
      for (int i = 0; i < 4; i++) {
        a[i] = rot[i][0] * edge4[n].a1 + rot[i][1] * edge4[n].a2 +
            rot[i][2] * edge4[n].a3 + rot[i][3] * edge4[n].a4;
        b[i] = rot[i][0] * edge4[n].b1 + rot[i][1] * edge4[n].b2 +
            rot[i][2] * edge4[n].b3 + rot[i][3] * edge4[n].b4;
      }
      edge4[n] = new Edge(a[0], a[1], a[2], a[3], b[0], b[1], b[2], b[3]);
    }
  }

  /**
   * Updates the 3D projection of the hypercube according to actual rotation angles of the
   * hypercube in 4D space.
   */
  public void update3DProjection() {
    double p1, p2, p3, q1, q2, q3, v1, v2, v3;

    //edges in direction x1
    for (int i = 0; i < 8; i++) {
      p1 = edge1[i].a1 * d / (d - edge1[i].a4);
      p2 = edge1[i].a2 * d / (d - edge1[i].a4);
      p3 = edge1[i].a3 * d / (d - edge1[i].a4);

      q1 = edge1[i].b1 * d / (d - edge1[i].b4);
      q2 = edge1[i].b2 * d / (d - edge1[i].b4);
      q3 = edge1[i].b3 * d / (d - edge1[i].b4);

      v1 = q1 - p1;
      v2 = q2 - p2;
      v3 = q3 - p3;
      e1[i].setXYZ(p1, p2, p3);
      e1[i].setNormalVector(new Vector3d(v1, v2, v3));
      e1[i].setMagnitude(Math.sqrt(v1 * v1 + v2 * v2 + v3 * v3));
    }

    //edges in direction x2
    for (int i = 0; i < 8; i++) {
      p1 = edge2[i].a1 * d / (d - edge2[i].a4);
      p2 = edge2[i].a2 * d / (d - edge2[i].a4);
      p3 = edge2[i].a3 * d / (d - edge2[i].a4);

      q1 = edge2[i].b1 * d / (d - edge2[i].b4);
      q2 = edge2[i].b2 * d / (d - edge2[i].b4);
      q3 = edge2[i].b3 * d / (d - edge2[i].b4);

      v1 = q1 - p1;
      v2 = q2 - p2;
      v3 = q3 - p3;

      e2[i].setXYZ(p1, p2, p3);
      e2[i].setNormalVector(new Vector3d(v1, v2, v3));
      e2[i].setMagnitude(Math.sqrt(v1 * v1 + v2 * v2 + v3 * v3));
    }

    //edges in direction x3
    for (int i = 0; i < 8; i++) {
      p1 = edge3[i].a1 * d / (d - edge3[i].a4);
      p2 = edge3[i].a2 * d / (d - edge3[i].a4);
      p3 = edge3[i].a3 * d / (d - edge3[i].a4);

      q1 = edge3[i].b1 * d / (d - edge3[i].b4);
      q2 = edge3[i].b2 * d / (d - edge3[i].b4);
      q3 = edge3[i].b3 * d / (d - edge3[i].b4);

      v1 = q1 - p1;
      v2 = q2 - p2;
      v3 = q3 - p3;

      e3[i].setXYZ(p1, p2, p3);
      e3[i].setNormalVector(new Vector3d(v1, v2, v3));
      e3[i].setMagnitude(Math.sqrt(v1 * v1 + v2 * v2 + v3 * v3));
    }

    //edges in direction x4
    for (int i = 0; i < 8; i++) {
      p1 = edge4[i].a1 * d / (d - edge4[i].a4);
      p2 = edge4[i].a2 * d / (d - edge4[i].a4);
      p3 = edge4[i].a3 * d / (d - edge4[i].a4);

      q1 = edge4[i].b1 * d / (d - edge4[i].b4);
      q2 = edge4[i].b2 * d / (d - edge4[i].b4);
      q3 = edge4[i].b3 * d / (d - edge4[i].b4);

      v1 = q1 - p1;
      v2 = q2 - p2;
      v3 = q3 - p3;

      e4[i].setXYZ(p1, p2, p3);
      e4[i].setNormalVector(new Vector3d(v1, v2, v3));
      e4[i].setMagnitude(Math.sqrt(v1 * v1 + v2 * v2 + v3 * v3));
    }

  }

  /**
   * Initializes the 3D scene. Creates edges array of projections of edges of hypercube
   * and adds them to the panel.
   */
  public void initialize3DProjection() {
    panel.clear();

    //edges in directions x1, x2, x3, x4
    for (int i = 0; i < 8; i++) {
      e1[i] = new DVectorLine(new Point3d(0, 0, 0), new Vector3d(1, 1, 1), w,
                              Color.red);
      panel.addDrawable3D(e1[i]);
      e2[i] = new DVectorLine(new Point3d(0, 0, 0), new Vector3d(1, 1, 1), w,
                              Color.green);
      panel.addDrawable3D(e2[i]);
      e3[i] = new DVectorLine(new Point3d(0, 0, 0), new Vector3d(1, 1, 1), w,
                              Color.blue);
      panel.addDrawable3D(e3[i]);
      e4[i] = new DVectorLine(new Point3d(0, 0, 0), new Vector3d(1, 1, 1), w,
                              Color.yellow);
      panel.addDrawable3D(e4[i]);
    }
    panel.setScenePEuler(1.3581, 0.3560, -1.7533);
  }

  /**
   * Muliplies 4D-matrix and a 4D-vector.
   */
  public double[] multiplyMatrixVector(double[][] a, double[] v) {
    double[] temp = new double[4];
    double sum;

    for (int i = 0; i < 4; i++) {
      sum = 0;
      for (int k = 0; k < 4; k++) {
        sum += a[i][k] * v[k];
      }
      temp[i] = sum;
    }

    return temp;
  }

  /**
   * Multiplies two matrices.
   */
  public double[][] multiplyMatrices(double[][] a, double[][] b) {
    double[][] temp = new double[4][4];
    double sum;
    for (int i = 0; i < 4; i++) {
      for (int j = 0; j < 4; j++) {
        sum = 0;
        for (int k = 0; k < 4; k++) {
          sum += a[i][k] * b[k][j];
        }
        temp[i][j] = sum;
      }
    }

    return temp;
  }

}