package org.opensourcephysics.tuleja.orbits;

import org.opensourcephysics.numerics.*;

// (c) Slavomir Tuleja, November 5, 2002
// Revised March 15, 2003
// The applet simulates motion of a satellite on its orbit around Earth

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;

public class JOrbitLabOutline
    extends JApplet
    implements Runnable, ActionListener, MouseMotionListener, MouseListener {

  ImageIcon iconPlay, iconStop, iconStepForward, iconStepBack, iconZoomIn, iconZoomOut;
  Image imMap;

  public BufferedImage image;

  private Graphics2D g2;
  private JButton btnStartStop, btnSet, btnForward, btnBack, btnZoomIn,
      btnZoomOut, btnState, btnReset;
  private JPanel pnlButtons, pnlLabels, pnlSouth;
  private JLabel lblTime, lblH, lblLatitude, lblLongitude;
  private Canvas canvas;
  private TextualInputWindow windowSettings;
  private Container cp;
  private Scale sc, scM;

  //...objects used for transformation between real-world coordinates and pixels
  private Thread thread;

  private OrbitODE ode = new Orbit();
  public ODEAdaptiveSolver odeAdaptiveSolver;

  private PixelGrabber pg; //object enabling pixel acces.
  private int[] pixelsM; //one-dimensional array of pixel values
  private boolean[][] pixB; //two-dimensional array corresponding to a map. True means land, false water.
  private boolean[][] visited; // boolean two-dimensional array. True - the pixel was added to an outline; Fase - otherwise
  private Outline[] outline = new Outline[500];
  private int numOfOutlines = 0;

  private double t, phi, theta, r;
  private double vR, vTheta, vPhi;
  private double rd, thetad, phid;
  public double phiInitial, thetaInitial, rInitial, vRInitial, vThetaInitial,
      vPhiInitial, dtInitial;
  public double dt;
  private final double Rearth = 6378e3;
  private double thetaOld, phiOld, rOld;
  private double xOld, yOld, zOld, x, y, z;
  private boolean end, impact, cursorOver = false;
  private int w = 450, h = 450; //image dimansions in pixels
  private int wM, hM; //width and height of the gif image containing the Earth map
  private int ci, cj; //coordinates of the cursor above canvas
  private double thetaS;

  private double th; //viewing angles definitions
  private double ph;
  private double ph0, th0;
  private int mouseX, mouseY, mouseX0, mouseY0;
  private double mouseConstant = 0.01;
  private int orbLength = 1000;
  private Point3D[] orbit = new Point3D[orbLength];
  private ProjPoint[] orbProj = new ProjPoint[orbLength];
  private double zoom;
  public String parH, parPhi, parLam, parVR, parVNorth, parVEast, parDT,
      parPhiSun, parZoom, parThView, parPhView;

  public static void main(String[] args) {
    JFrame fr = new JFrame("JOrbitLabOutline");
    fr.setSize(450, 550);
    JOrbitLabOutline applet = new JOrbitLabOutline();
    applet.initApp();
    applet.start();
    fr.getContentPane().add(applet, BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }

  public void loadParameters() {
    //loading the applet parameters
    parH = getParameter("h");
    parPhi = getParameter("phi");
    parLam = getParameter("lam");
    parVR = getParameter("vUp");
    parVNorth = getParameter("vNorth");
    parVEast = getParameter("vEast");
    parDT = getParameter("dt");
    parPhiSun = getParameter("phiSun");
    parZoom = getParameter("zoom");
    parThView = getParameter("thetaView");
    parPhView = getParameter("phiView");
  }

  /** Returns an ImageIcon, or null if the path was invalid. */
  protected static ImageIcon createImageIcon(String path) {
    java.net.URL imgURL = JOrbitLabOutline.class.getResource(path);
    if (imgURL != null) {
      return new ImageIcon(imgURL);
    }
    else {
      System.err.println("Couldn't find file: " + path);
      return null;
    }
  }


  public void createGUI() {

    String imagePath = "/org/opensourcephysics/tuleja/images/";
    iconPlay = createImageIcon(imagePath+"Play.gif");
    iconStop = createImageIcon(imagePath+"Stop.gif");
    iconStepForward = createImageIcon(imagePath+"StepForward.gif");
    iconStepBack = createImageIcon(imagePath+"StepBack.gif");
    iconZoomIn = createImageIcon(imagePath+"ZoomIn.gif");
    iconZoomOut = createImageIcon(imagePath+"ZoomOut.gif");

    imMap = createImageIcon(imagePath+"map.gif").getImage();


    //creating GUI
    cp = getContentPane();
    image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
    canvas = new Canvas(image, this);

    wM = imMap.getWidth(null);
    hM = imMap.getHeight(null);

    // We create a two-dimensional array of integers and grab pixel values from imMap
    // to it, using the PixelGrabber.
    pixelsM = new int[wM * hM];
    pg = new PixelGrabber(imMap, 0, 0, wM, hM, pixelsM, 0, wM);

    try {
      pg.grabPixels();
    }
    catch (InterruptedException e) {
      System.err.println("Interrupted waiting for pixels!");
      return;
    }

    if ( (pg.getStatus() & ImageObserver.ABORT) != 0) {
      System.err.println("Image fetch aborted or errored");
      return;
    }

    pixB = new boolean[wM][hM];
    visited = new boolean[wM][hM];
    for (int i = 0; i < wM; i++) {
      for (int j = 0; j < hM; j++) {
        if (pixelsM[wM * j + i] == -328147)
          pixB[i][j] = true;
        else
          pixB[i][j] = false;
        visited[i][j] = false;
      }
    }

    findOutlines();

    btnStartStop = new JButton("Start", iconPlay);
    btnSet = new JButton("Settings");
    btnForward = new JButton("Step", iconStepForward);
    btnBack = new JButton("Step", iconStepBack);
    btnZoomIn = new JButton("Zoom in", iconZoomIn);
    btnZoomOut = new JButton("Zoom out", iconZoomOut);
    btnReset = new JButton("Reset");
    btnState = new JButton("State");

    lblTime = new JLabel("t = 0 s");
    lblH = new JLabel("h = ".concat(MyUtils.roundOff(Double.valueOf(parH).
        doubleValue() / 1000.0, 1)).concat(" km"));
    lblLatitude = new JLabel("lat.: ");
    lblLongitude = new JLabel("long.: ");

    pnlButtons = new JPanel();
    pnlButtons.setLayout(new GridLayout(2, 4));
    pnlButtons.add(btnStartStop);
    pnlButtons.add(btnBack);
    pnlButtons.add(btnForward);
    pnlButtons.add(btnReset);
    pnlButtons.add(btnZoomIn);
    pnlButtons.add(btnZoomOut);
    pnlButtons.add(btnState);
    pnlButtons.add(btnSet);

    pnlLabels = new JPanel();
    pnlLabels.setLayout(new GridLayout(1, 4));
    pnlLabels.add(lblTime);
    pnlLabels.add(lblH);
    pnlLabels.add(lblLatitude);
    pnlLabels.add(lblLongitude);

    pnlSouth = new JPanel();
    pnlSouth.setLayout(new BoxLayout(pnlSouth, BoxLayout.Y_AXIS));
    pnlSouth.add(pnlLabels);
    pnlSouth.add(pnlButtons);

    cp.setLayout(new BorderLayout());
    cp.add(canvas, BorderLayout.CENTER);
    cp.add(pnlSouth, BorderLayout.SOUTH);

    btnStartStop.addActionListener(this);
    btnForward.addActionListener(new StepForwardHandler(this));
    btnBack.addActionListener(new StepBackHandler(this));
    btnSet.addActionListener(this);
    btnZoomIn.addActionListener(this);
    btnZoomOut.addActionListener(this);
    btnReset.addActionListener(this);
    btnState.addActionListener(this);

    g2 = image.createGraphics();

    canvas.addMouseMotionListener(this);
    canvas.addMouseListener(this);

    windowSettings = new TextualInputWindow(this);

    rInitial = Double.valueOf(parH).doubleValue() + Rearth;
    thetaInitial = 90 - Double.valueOf(parPhi).doubleValue();
    phiInitial = toLongitude(Double.valueOf(parLam).doubleValue() / 180.0 *
                             Math.PI);
    vRInitial = Double.valueOf(parVR).doubleValue();
    vThetaInitial = -Double.valueOf(parVNorth).doubleValue();
    vPhiInitial = Double.valueOf(parVEast).doubleValue();
    dtInitial = Double.valueOf(parDT).doubleValue();
    thetaS = Math.PI / 2.0 -
        Double.valueOf(parPhiSun).doubleValue() / 180.0 * Math.PI;
    zoom = Double.valueOf(parZoom).doubleValue();
    th = Double.valueOf(parThView).doubleValue() / 180.0 * Math.PI;
    ph = Double.valueOf(parPhView).doubleValue() / 180.0 * Math.PI;

    setScales();
    reset();
    setSimulation(rInitial, thetaInitial, phiInitial, vRInitial,
                  vThetaInitial, vPhiInitial, 0, dtInitial);

  }

  public void init() {
    loadParameters();
    createGUI();
  }

  public void initApp() {
    parH = "122000";
    parPhi = "28.4";
    parLam = "-80.6";
    parVR = "0";
    parVNorth = "6784";
    parVEast = "3917";
    parDT = "30";
    parPhiSun = "-18";
    parZoom = "1";
    parThView = "40";
    parPhView = "20";

    createGUI();
  }

  public void reset() {
    setSimulation(rInitial, thetaInitial, phiInitial, vRInitial,
                  vThetaInitial, vPhiInitial, 0, dtInitial);
  }

  public void setSimulation(double r, double theta, double phi, double vR,
                            double vTheta, double vPhi, double t, double dt) {
    //angles in DEGREES!!!
    end = false;
    impact = false;
    this.t = t;
    this.r = r;
    this.theta = theta / 180.0 * Math.PI;
    this.phi = phi / 180.0 * Math.PI;
    this.dt = dt;

    // InitialConditions are set here
    ode.setIC(r, theta / 180.0 * Math.PI, phi / 180.0 * Math.PI, vR, vTheta,
              vPhi);

    lblTime.setText("t = ".concat(MyUtils.roundOff(t, 0)).concat(" s"));
    lblH.setText("h = ".concat(MyUtils.roundOff( (r - Rearth) / 1000.0, 1)).
                 concat(
        " km"));
    lblLatitude.setText("lat.: ".concat(MyUtils.roundOff(toLatitude(this.theta),
        1)).
                        concat(" "));
    lblLongitude.setText("long.: ".concat(MyUtils.roundOff(toLongitude(this.phi),
        1)).
                         concat(" "));

    //inicialization of ODESolver
    odeAdaptiveSolver = new RK45MultiStep(ode);
    odeAdaptiveSolver.initialize(dt);
    odeAdaptiveSolver.setTolerance(1e-6);
    odeAdaptiveSolver.setStepSize(dt);

    for (int i = 0; i < orbLength; i++) {
      orbit[i] = new Point3D(ode.getX() / Rearth, ode.getY() / Rearth,
                             ode.getZ() / Rearth);
      orbProj[i] = new ProjPoint(ode.getPhi(), ode.getTheta());
    }

    drawBottomLayer(g2);
    canvas.repaint();
  }

  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Stop")) {
      //after pressing STOP
      end = true;
      btnStartStop.setText("Start");
      btnStartStop.setIcon(iconPlay);
      btnForward.setEnabled(true);
      btnBack.setEnabled(true);
      btnSet.setEnabled(true);
      btnReset.setEnabled(true);
      btnState.setEnabled(true);
    }
    else if (e.getActionCommand().equals("Start") && (!impact)) {
      //after pressing START
      end = false;
      btnStartStop.setText("Stop");
      btnStartStop.setIcon(iconStop);
      btnForward.setEnabled(false);
      btnBack.setEnabled(false);
      btnSet.setEnabled(false);
      btnReset.setEnabled(false);
      btnState.setEnabled(false);
      thread = new Thread(this);
      thread.start();
    }

    else if (e.getActionCommand().equals("Step")) {
    }
    else if (e.getActionCommand().equals("Reset")) {
      reset();
      repaint();
    }
    else if (e.getActionCommand().equals("State")) {
      printOutState();
    }
    else if (e.getActionCommand().equals("Settings")) {
      //after pressing Settings
      windowSettings.setLocation();
      windowSettings.show();
    }
    else if (e.getActionCommand().equals("Zoom in")) {
      zoom /= 1.25;
      if (zoom <= 0.8) {
        zoom = 0.8;
        btnZoomIn.setEnabled(false);
      }
      setScales();
    }
    else if (e.getActionCommand().equals("Zoom out")) {
      btnZoomIn.setEnabled(true);
      zoom *= 1.25;
      setScales();
    }
  }

  public void printOutState() {
    String s;
    s = new String("State of the satellite:\n");
    s = s.concat("Height:  ").concat(MyUtils.roundOff(ode.getR() - Rearth, 0)).
        concat(" m\n");
    s = s.concat("Latitude:  ").concat(MyUtils.roundOff(toLatitude(ode.getTheta()),
        1)).concat(" \u00b0 \n");
    s = s.concat("Longitude:  ").concat(MyUtils.roundOff(toLongitude(ode.getPhi() -
        7.272e-5 * ode.getT()), 1)).concat(" \u00b0 \n");
    s = s.concat("Upward velocity:  ").concat(MyUtils.roundOff(ode.getVR(), 1)).
        concat(" m/s \n");
    s = s.concat("Northward velocity:  ").concat(MyUtils.roundOff( -ode.
        getVTheta(), 1)).concat(" m/s \n");
    s = s.concat("Eastward velocity:  ").concat(MyUtils.roundOff(ode.getVPhi(),
        1)).concat(" m/s \n");
    s = s.concat("Velocity magnitude:  ").concat(MyUtils.roundOff(Math.sqrt(ode.
        getVR() * ode.getVR() + ode.getVPhi() * ode.getVPhi() +
        ode.getVTheta() * ode.getVTheta()),
        1)).concat(" m/s \n");

    JOptionPane.showMessageDialog(this, s, "State of the satellite",
                                  JOptionPane.INFORMATION_MESSAGE);
  }

  private void findOutlines() {
    int i;
    int j;
    numOfOutlines = 0;
    for (i = 0; i < wM; i++) {
      for (j = 1; j < hM - 1; j++) {
        if ( (!visited[i][j]) && (isOnEdge(new IntegerPoint(i, j)))) {
          outline[numOfOutlines] = new Outline();
          lookForOutline(outline[numOfOutlines], i, j);
          numOfOutlines++;
        }
      }
    }
  }

  private boolean isOnEdge(IntegerPoint p) {
    int i = p.getI();
    int j = p.getJ();
    boolean b = (!pixB[getIMinus(i)][j - 1]) | (!pixB[getIMinus(i)][j]) |
        (!pixB[getIMinus(i)][j + 1]) | (!pixB[i][j - 1]) | (!pixB[i][j + 1]) |
        (!pixB[getIPlus(i)][j - 1]) | (!pixB[getIPlus(i)][j]) |
        (!pixB[getIPlus(i)][j + 1]);
    b &= pixB[i][j];
    return b;
  }

  /**
   * Connects opposite sides of the image
   */
  private int getIPlus(int i) {
    if ( (i + 1) < wM)
      return i + 1;
    else
      return 0;
  }

  private int getIMinus(int i) {
    if ( (i - 1) >= 0)
      return i - 1;
    else
      return wM - 1;
  }

  /**
   * To a given pixel [i,j] finds an outline going through it
   */
  private void lookForOutline(Outline outline, int i, int j) {
    visited[i][j] = true;
    outline.addPoint(new IntegerPoint(i, j));
    if ( (!visited[i][j + 1]) && (isOnEdge(new IntegerPoint(i, j + 1))))
      lookForOutline(outline, i, j + 1);
    if ( (!visited[i][j - 1]) && (isOnEdge(new IntegerPoint(i, j - 1))))
      lookForOutline(outline, i, j - 1);
    if ( (!visited[getIPlus(i)][j - 1]) &&
        (isOnEdge(new IntegerPoint(getIPlus(i), j - 1))))
      lookForOutline(outline, getIPlus(i), j - 1);
    if ( (!visited[getIPlus(i)][j]) &&
        (isOnEdge(new IntegerPoint(getIPlus(i), j))))
      lookForOutline(outline, getIPlus(i), j);
    if ( (!visited[getIPlus(i)][j + 1]) &&
        (isOnEdge(new IntegerPoint(getIPlus(i), j + 1))))
      lookForOutline(outline, getIPlus(i), j + 1);
    if ( (!visited[getIMinus(i)][j - 1]) &&
        (isOnEdge(new IntegerPoint(getIMinus(i), j - 1))))
      lookForOutline(outline, getIMinus(i), j - 1);
    if ( (!visited[getIMinus(i)][j]) &&
        (isOnEdge(new IntegerPoint(getIMinus(i), j))))
      lookForOutline(outline, getIMinus(i), j);
    if ( (!visited[getIMinus(i)][j + 1]) &&
        (isOnEdge(new IntegerPoint(getIMinus(i), j + 1))))
      lookForOutline(outline, getIMinus(i), j + 1);
  }

  public void drawOutlines(Graphics2D g2) {
    int i1, j1, i2, j2;
    double x1, x2, y1, y2, z1, z2, sx, sy, sz, scs;
    double ph1, ph2, th1, th2;
    int iP1, jP1, iP2, jP2;
    g2.setColor(Color.blue);
    g2.fillOval(sc.xToi( -1), sc.yToj(1), sc.xToi(1) - sc.xToi( -1),
                sc.yToj( -1) - sc.yToj(1));
    g2.setColor(Color.green);
    for (int n = 0; n < numOfOutlines; n++) {
      for (int k = 0; k < outline[n].n - 1; k++) {
        i1 = outline[n].getPoint(k).getI();
        j1 = outline[n].getPoint(k).getJ();
        i2 = outline[n].getPoint(k + 1).getI();
        j2 = outline[n].getPoint(k + 1).getJ();
        ph1 = adjust(scM.iTox(i1)); //rotacia zeme
        th1 = scM.jToy(j1);
        ph2 = adjust(scM.iTox(i2));
        th2 = scM.jToy(j2);
        x1 = Math.cos(ph1) * Math.sin(th1);
        y1 = Math.sin(ph1) * Math.sin(th1);
        z1 = Math.cos(th1);
        x2 = Math.cos(ph2) * Math.sin(th2);
        y2 = Math.sin(ph2) * Math.sin(th2);
        z2 = Math.cos(th2);
        sx = 0.5 * (x2 + x1);
        sy = 0.5 * (y2 + y1);
        sz = 0.5 * (z2 + z1);
        scs = sx * Math.cos(ph) * Math.sin(th) +
            sy * Math.sin(ph) * Math.sin(th) + sz * Math.cos(th);
        iP1 = sc.xToi(getXP(x1, y1, z1));
        jP1 = sc.yToj(getYP(x1, y1, z1));
        iP2 = sc.xToi(getXP(x2, y2, z2));
        jP2 = sc.yToj(getYP(x2, y2, z2));

        if ( (Math.abs(i1 - i2) < 3) && (Math.abs(j1 - j2) < 3) && (scs > 0))
          g2.drawLine(iP1, jP1, iP2, jP2);
      }
    }
  }

  public void drawTopLayer(Graphics2D g2) {
    drawOutlines(g2);
    drawShaddow(g2);
    drawVisibilityArea(g2);
    drawSatellite(g2);
  }

  private void drawSatellite(Graphics2D g2) {
    double x1, y1, z1, x2, y2, z2, xP1, yP1, xP2, yP2, xPs, yPs, xs, ys, zs;
    double scs, dis;
    double ph1, ph2, th1, th2;

    //drawing the projection of the trajectory
    g2.setColor(Color.yellow);
    for (int i = 0; i < orbLength - 1; i++) {
      ph1 = adjust(orbProj[i].getPhi());
      th1 = orbProj[i].getTheta();
      ph2 = adjust(orbProj[i + 1].getPhi());
      th2 = orbProj[i + 1].getTheta();
      x1 = Math.cos(ph1) * Math.sin(th1);
      y1 = Math.sin(ph1) * Math.sin(th1);
      z1 = Math.cos(th1);
      x2 = Math.cos(ph2) * Math.sin(th2);
      y2 = Math.sin(ph2) * Math.sin(th2);
      z2 = Math.cos(th2);

      xP1 = getXP(x1, y1, z1);
      yP1 = getYP(x1, y1, z1);
      xP2 = getXP(x2, y2, z2);
      yP2 = getYP(x2, y2, z2);
      xPs = 0.5 * (xP1 + xP2);
      yPs = 0.5 * (yP1 + yP2);
      xs = 0.5 * (x1 + x2);
      ys = 0.5 * (y1 + y2);
      zs = 0.5 * (z1 + z2);
      dis = Math.sqrt(xPs * xPs + yPs * yPs);
      scs = Math.cos(ph) * Math.sin(th) * xs + Math.sin(ph) * Math.sin(th) * ys +
          Math.cos(th) * zs;
      if (! (dis < 1) || ! (scs < 0)) {
        g2.drawLine(sc.xToi(xP1), sc.yToj(yP1), sc.xToi(xP2), sc.yToj(yP2));
      }
    }
    g2.setColor(Color.red);
    //drawing trajectory
    for (int i = 0; i < orbLength - 1; i++) {
      x1 = orbit[i].getX();
      y1 = orbit[i].getY();
      z1 = orbit[i].getZ();
      x2 = orbit[i + 1].getX();
      y2 = orbit[i + 1].getY();
      z2 = orbit[i + 1].getZ();

      xP1 = getXP(x1, y1, z1);
      yP1 = getYP(x1, y1, z1);
      xP2 = getXP(x2, y2, z2);
      yP2 = getYP(x2, y2, z2);
      xPs = 0.5 * (xP1 + xP2);
      yPs = 0.5 * (yP1 + yP2);
      xs = 0.5 * (x1 + x2);
      ys = 0.5 * (y1 + y2);
      zs = 0.5 * (z1 + z2);
      dis = Math.sqrt(xPs * xPs + yPs * yPs);
      scs = Math.cos(ph) * Math.sin(th) * xs + Math.sin(ph) * Math.sin(th) * ys +
          Math.cos(th) * zs;
      if (! (dis < 1) || ! (scs < 0)) {
        g2.drawLine(sc.xToi(xP1), sc.yToj(yP1), sc.xToi(xP2), sc.yToj(yP2));
        if (i == orbLength - 2) {
          g2.setColor(Color.red);
          g2.fillOval(sc.xToi(getXP(ode.getX() / Rearth, ode.getY() / Rearth,
                                    ode.getZ() / Rearth)) - 3,
                      sc.yToj(getYP(ode.getX() / Rearth, ode.getY() / Rearth,
                                    ode.getZ() / Rearth)) - 3, 6, 6);
          g2.setColor(Color.black);
          g2.drawOval(sc.xToi(getXP(ode.getX() / Rearth, ode.getY() / Rearth,
                                    ode.getZ() / Rearth)) - 3,
                      sc.yToj(getYP(ode.getX() / Rearth, ode.getY() / Rearth,
                                    ode.getZ() / Rearth)) - 3, 6, 6);
        }
      }
    }

    //box around the image
    g2.setColor(Color.black);
    g2.drawLine(0, 0, w - 1, 0);
    g2.drawLine(w - 1, 0, w - 1, h - 1);
    g2.drawLine(w - 1, h - 1, 0, h - 1);
    g2.drawLine(0, h - 1, 0, 0);
    repaint();
  }

  private void drawShaddow(Graphics2D g2) {
    double nSx, nSy, nSz; //vector pointing from Earth center to the Sun
    double px, py, rs, tt, x, y, z;
    double phiS = 0;
    while (phiS > Math.PI)
      phiS -= 2 * Math.PI;
    while (phiS < -Math.PI)
      phiS += 2 * Math.PI;
    int di = 2, dj = 2;
    nSx = Math.sin(thetaS) * Math.cos(phiS);
    nSy = Math.sin(thetaS) * Math.sin(phiS);
    nSz = Math.cos(thetaS);
    //drawing shaddow
    for (int j = 0; j < h; j += 2 * dj) {
      for (int i = 0; i < w - di; i += di) {
        px = sc.iTox(i);
        py = sc.jToy(j);
        rs = 1 - px * px - py * py;
        if (rs >= 0) {
          tt = Math.sqrt(rs);
          x = -px * Math.sin(ph) - py * Math.cos(ph) * Math.cos(th) +
              tt * Math.cos(ph) * Math.sin(th);
          y = px * Math.cos(ph) - py * Math.sin(ph) * Math.cos(th) +
              tt * Math.sin(ph) * Math.sin(th);
          z = py * Math.sin(th) + tt * Math.cos(th);

          if (x * nSx + y * nSy + z * nSz < 0) {
            g2.setColor(Color.black);
            g2.drawLine(i, j, i, j);
          }

          px = sc.iTox(i + di / 2);
          py = sc.jToy(j + dj);
          rs = 1 - px * px - py * py;
          if (rs >= 0) {
            tt = Math.sqrt(rs);
            x = -px * Math.sin(ph) - py * Math.cos(ph) * Math.cos(th) +
                tt * Math.cos(ph) * Math.sin(th);
            y = px * Math.cos(ph) - py * Math.sin(ph) * Math.cos(th) +
                tt * Math.sin(ph) * Math.sin(th);
            z = py * Math.sin(th) + tt * Math.cos(th);

            if (x * nSx + y * nSy + z * nSz < 0) {
              g2.setColor(Color.black);
              g2.drawLine(i + di / 2, j + dj, i + di / 2, j + dj);
            }
          }
        }
      }
    }
  }

  private void drawVisibilityArea(Graphics2D g2) {
    double thetaP, alphaP; //coordinates of a point on the edge of the visibility area circle
    double thetaPOld, alphaPOld;
    double rho = Math.acos(Rearth / ode.getR()); //radius of visibility circle as an angle
    double beta = 0, dBeta = Math.PI / 90.0;
    double ph1, ph2, th1, th2, ph3, ph4, th3, th4;
    int i1, i2, j1, j2, i3, i4, j3, j4;
    double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4;
    double nx, ny, nz;

    nx = Math.sin(th) * Math.cos(ph);
    ny = Math.sin(th) * Math.sin(ph);
    nz = Math.cos(th);

    g2.setColor(Color.white);

    thetaPOld = Math.acos(Math.cos(ode.getTheta()) * Math.cos(rho) +
                          Math.sin(ode.getTheta()) * Math.sin(rho) *
                          Math.cos(beta));
    alphaPOld = Math.asin(Math.sin(beta) * Math.sin(rho) /
                          Math.sin(thetaPOld));
    if (Math.cos(rho) - Math.cos(thetaPOld) * Math.cos(ode.getTheta()) < 0)
      alphaPOld = Math.PI - alphaPOld;

    while (beta < Math.PI) {
      beta += dBeta;

      thetaP = Math.acos(Math.cos(ode.getTheta()) * Math.cos(rho) +
                         Math.sin(ode.getTheta()) * Math.sin(rho) *
                         Math.cos(beta));
      alphaP = Math.asin(Math.sin(beta) * Math.sin(rho) / Math.sin(thetaP));
      if (Math.cos(rho) - Math.cos(thetaP) * Math.cos(ode.getTheta()) < 0)
        alphaP = Math.PI - alphaP;

      ph1 = ode.getPhi() - alphaPOld;
      ph2 = ode.getPhi() - alphaP;
      th1 = thetaPOld;
      th2 = thetaP;
      ph3 = ode.getPhi() + alphaPOld;
      ph4 = ode.getPhi() + alphaP;
      th3 = thetaPOld;
      th4 = thetaP;

      x1 = Math.cos(ph1) * Math.sin(th1);
      y1 = Math.sin(ph1) * Math.sin(th1);
      z1 = Math.cos(th1);
      x2 = Math.cos(ph2) * Math.sin(th2);
      y2 = Math.sin(ph2) * Math.sin(th2);
      z2 = Math.cos(th2);
      x3 = Math.cos(ph3) * Math.sin(th3);
      y3 = Math.sin(ph3) * Math.sin(th3);
      z3 = Math.cos(th3);
      x4 = Math.cos(ph4) * Math.sin(th4);
      y4 = Math.sin(ph4) * Math.sin(th4);
      z4 = Math.cos(th4);

      i1 = sc.xToi(getXP(x1, y1, z1));
      j1 = sc.yToj(getYP(x1, y1, z1));
      i2 = sc.xToi(getXP(x2, y2, z2));
      j2 = sc.yToj(getYP(x2, y2, z2));
      i3 = sc.xToi(getXP(x3, y3, z3));
      j3 = sc.yToj(getYP(x3, y3, z3));
      i4 = sc.xToi(getXP(x4, y4, z4));
      j4 = sc.yToj(getYP(x4, y4, z4));

      if (0.5 * (x1 + x2) * nx + 0.5 * (y1 + y2) * ny + 0.5 * (z1 + z2) * nz >
          0)
        g2.drawLine(i1, j1, i2, j2);
      if (0.5 * (x3 + x4) * nx + 0.5 * (y3 + y4) * ny + 0.5 * (z3 + z4) * nz >
          0)
        g2.drawLine(i3, j3, i4, j4);

      thetaPOld = thetaP;
      alphaPOld = alphaP;
    }
  }

  public double adjust(double phi) {
    // first adjusts phi with respect to Earth rotation and then
    // transforms it into inteval from -PI to PI
    double phiA = phi + 7.272e-5 * t;
    while (phiA > Math.PI)
      phiA -= 2 * Math.PI;
    while (phiA < -Math.PI)
      phiA += 2 * Math.PI;
    return phiA;
  }

  public void drawBottomLayer(Graphics2D g2) {
    g2.setColor(Color.black);
    g2.fillRect(0, 0, w, h);
    g2.setColor(Color.black);
    g2.drawRect(0, 0, w - 1, h - 1);
  }

  public void setScales() {
    sc = new Scale( -1.05 * zoom, 1.05 * zoom, -1.05 * zoom, 1.05 * zoom, 0,
                   w - 1, 0, h - 1);
    scM = new Scale( -Math.PI, Math.PI, Math.PI, 0, 0, wM - 1, 0, hM - 1);
  }

  /**
       * Returns x-coordinate of a projection of a 3-d point to a plane perpendicular
   * to a viewing vector.
   */
  public double getXP(double x, double y, double z) {
    return -x * Math.sin(ph) + y * Math.cos(ph);
  }

  /**
       * Returns y-coordinate of a projection of a 3-d point to a plane perpendicular
   * to a viewing vector.
   */
  public double getYP(double x, double y, double z) {
    return -x * Math.cos(ph) * Math.cos(th) - y * Math.sin(ph) * Math.cos(th) +
        z * Math.sin(th);
  }

  public void run() {
    while (!end && !impact) {
      simulationStep();
      try {
        thread.sleep(100);
      }
      catch (InterruptedException exception) {
        System.err.println("Interrupted...");
      }
    }
  }

  public void simulationStep() {
    phiOld = ode.getPhi();
    thetaOld = ode.getTheta();
    rOld = ode.getR();
    for (int i = 0; i < orbLength - 1; i++) {
      orbit[i] = orbit[i + 1];
      orbProj[i] = orbProj[i + 1];
    }
    t += odeAdaptiveSolver.step();

    orbit[orbLength -
        1] = new Point3D(ode.getX() / Rearth, ode.getY() / Rearth,
                         ode.getZ() / Rearth);
    orbProj[orbLength -
        1] = new ProjPoint(ode.getPhi() - 7.272e-5 * t, ode.getTheta());
    phi = ode.getPhi();
    theta = ode.getTheta();
    r = ode.getR();

    lblTime.setText("t = ".concat(MyUtils.roundOff(t, 0)).concat(" s"));
    lblH.setText("h = ".concat(MyUtils.roundOff( (r - Rearth) / 1000.0, 1)).
                 concat(
        " km"));
    lblLatitude.setText("lat.: ".concat(MyUtils.roundOff(toLatitude(theta), 1)).
                        concat(" "));
    lblLongitude.setText("long.: ".concat(MyUtils.roundOff(toLongitude(phi -
        7.272e-5 * t), 1)).concat(" "));

    if (r < Rearth) { //osetrenie impactu na zem
      end = true;
      impact = true;
      btnStartStop.setText("Start");
      btnStartStop.setIcon(iconPlay);
      btnForward.setEnabled(true);
      btnBack.setEnabled(true);
      btnSet.setEnabled(true);
      btnReset.setEnabled(true);
      btnState.setEnabled(true);
    }
    if (impact)
      JOptionPane.showMessageDialog(this,
          "I am sorry but the satellite has hit the Earth...",
          "Warning", JOptionPane.WARNING_MESSAGE);
  }

  private double toLatitude(double theta) {
    //input theta in radians
    //latidude returned in degrees
    double th = theta / Math.PI * 180.0;
    return 90 - th;
  }

  private double toLongitude(double phi) {
    //input phi in radians
    //longitude returned in degrees
    double ph = phi / Math.PI * 180.0;
    while (ph > 180)
      ph = ph - 360;
    while (ph < -180)
      ph = ph + 360;
    return ph;
  }

  public void mousePressed(MouseEvent e) {
    //remembers the pixel above which the dragging has started
    th0 = th;
    ph0 = ph;
    mouseX0 = e.getX();
    mouseY0 = e.getY();
  }

  public void mouseDragged(MouseEvent e) {
    //used to change the viewing direction
    mouseX = e.getX();
    mouseY = e.getY();
    ph = ph0 - mouseConstant * (mouseX - mouseX0);
    ph = upravPh(ph);
    th = th0 - mouseConstant * (mouseY - mouseY0);
    if (th < 0) {
      th = 0;
    }
    if (th > Math.PI) {
      th = Math.PI;
    }
    canvas.repaint();
  }

  public void mouseReleased(MouseEvent e) {

  }

  public void mouseEntered(MouseEvent e) {
    cursorOver = true;
    canvas.repaint();
  }

  public void mouseExited(MouseEvent e) {
    cursorOver = false;
    canvas.repaint();
  }

  public void mouseClicked(MouseEvent e) {
  }

  public void mouseMoved(MouseEvent e) {
  }

  public double upravPh(double ph) {
    //sets the value of ph angle to interval <0,2PI>
    if (ph > 2 * Math.PI) {
      while (ph > 2 * Math.PI) {
        ph = ph - 2 * Math.PI;
      }
    }
    else
    if (ph < 0) {
      while (ph < 0) {
        ph = ph + 2 * Math.PI;
      }
    }
    return ph;
  }
}

class TextualInputWindow
    extends JDialog
    implements ActionListener {
  JOrbitLabOutline applet;
  JLabel lblH, lblLatitude, lblLongitude, lblVR, lblVNorth, lblVEast, lblDT;
  JTextField tfH, tfLatitude, tfLongitude, tfVR, tfVNorth, tfVEast, tfDT;
  JPanel pnlCenter, pnlButtons;
  JButton btnOK, btnCancel;
  String sH, sTheta, sPhi, sVR, sVTheta, sVPhi, sDT;

  //constructor
  public TextualInputWindow(JOrbitLabOutline applet) {
    setModal(true);
    setTitle("Simulation Settings");
    setSize(200, 340);
    setResizable(false);

    this.applet = applet;

    Container cp = getContentPane();
    btnOK = new JButton("OK");
    btnCancel = new JButton("Cancel");
    lblH = new JLabel("h / m:");
    lblLatitude = new JLabel("latitude / degrees:");
    lblLongitude = new JLabel("longitude / degrees:");
    lblVR = new JLabel("upward velocity / m/s:");
    lblVNorth = new JLabel("northward velocity / m/s:");
    lblVEast = new JLabel("eastward velocity / m/s:");
    lblDT = new JLabel("time step /s:");
    tfH = new JTextField(applet.parH);
    tfLatitude = new JTextField(applet.parPhi);
    tfLongitude = new JTextField(applet.parLam);
    tfVR = new JTextField(applet.parVR);
    tfVNorth = new JTextField(applet.parVNorth);
    tfVEast = new JTextField(applet.parVEast);
    tfDT = new JTextField(applet.parDT);

    pnlCenter = new JPanel();
    pnlCenter.setLayout(new BoxLayout(pnlCenter, BoxLayout.Y_AXIS));
    pnlCenter.add(lblH);
    pnlCenter.add(tfH);
    pnlCenter.add(lblLatitude);
    pnlCenter.add(tfLatitude);
    pnlCenter.add(lblLongitude);
    pnlCenter.add(tfLongitude);
    pnlCenter.add(lblVR);
    pnlCenter.add(tfVR);
    pnlCenter.add(lblVNorth);
    pnlCenter.add(tfVNorth);
    pnlCenter.add(lblVEast);
    pnlCenter.add(tfVEast);
    pnlCenter.add(lblDT);
    pnlCenter.add(tfDT);

    pnlCenter.setAlignmentX(Component.LEFT_ALIGNMENT);
    pnlCenter.setPreferredSize(new Dimension(150, 350));

    pnlCenter.add(Box.createVerticalGlue());
    pnlCenter.setBorder(BorderFactory.createEmptyBorder(0, 5, 0, 5));

    pnlButtons = new JPanel();
    pnlButtons.setLayout(new GridLayout(1, 2));
    pnlButtons.add(btnOK);
    pnlButtons.add(btnCancel);
    pnlButtons.setBorder(BorderFactory.createEmptyBorder(0, 5, 0, 5));

    cp.add(pnlButtons, BorderLayout.SOUTH);
    cp.add(pnlCenter, BorderLayout.CENTER);

    btnOK.addActionListener(this);
    btnCancel.addActionListener(this);
    //load old values
    sH = tfH.getText();
    sTheta = tfLatitude.getText();
    sPhi = tfLongitude.getText();
    sVR = tfVR.getText();
    sVTheta = tfVNorth.getText();
    sVPhi = tfVEast.getText();
    sDT = tfDT.getText();
  }

  public void setLocation() {
    //Sets the location of Textual Input Window so that it is
    //centered at the applet...
    Point cpP = applet.getContentPane().getLocationOnScreen();
    int ic = Math.round( (float) cpP.getX());
    int jc = Math.round( (float) cpP.getY());
    setLocation(new Point(ic + 125, jc + 75));
  }

  public void actionPerformed(ActionEvent e) {
    double r, theta, phi, vR, vTheta, vPhi, dt; //uhly v stupnoch!!!
    boolean allOK;
    if (e.getActionCommand().equals("OK")) {
      //after pressing OK button...
      allOK = true;
      try {
        r = Double.valueOf(tfH.getText()).doubleValue() + 6378e3;
        sH = tfH.getText();
        applet.rInitial = r;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfH.setText(sH);
        r = Double.valueOf(sH).doubleValue() + 6378e3;
        allOK = false;
      }
      try {
        theta = 90 - Double.valueOf(tfLatitude.getText()).doubleValue();
        sTheta = tfLatitude.getText();
        applet.thetaInitial = theta;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nplease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfLatitude.setText(sTheta);
        theta = 90 - Double.valueOf(sTheta).doubleValue();
        allOK = false;
      }
      try {
        phi = Double.valueOf(tfLongitude.getText()).doubleValue();
        if (phi < 0)
          phi = phi + 360;
        sPhi = tfLongitude.getText();
        applet.phiInitial = phi;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfLongitude.setText(sPhi);
        phi = Double.valueOf(sPhi).doubleValue();
        if (phi < 0)
          phi = phi + 360;
        allOK = false;
      }
      try {
        vR = Double.valueOf(tfVR.getText()).doubleValue();
        sVR = tfVR.getText();
        applet.vRInitial = vR;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfVR.setText(sVR);
        vR = Double.valueOf(sVR).doubleValue();
        allOK = false;
      }
      try {
        vTheta = -Double.valueOf(tfVNorth.getText()).doubleValue();
        sVTheta = tfVNorth.getText();
        applet.vThetaInitial = vTheta;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfVNorth.setText(sVTheta);
        vTheta = -Double.valueOf(sVTheta).doubleValue();
        allOK = false;
      }
      try {
        vPhi = Double.valueOf(tfVEast.getText()).doubleValue();
        sVPhi = tfVEast.getText();
        applet.vPhiInitial = vPhi;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfVEast.setText(sVPhi);
        vPhi = Double.valueOf(sVPhi).doubleValue();
        allOK = false;
      }
      try {
        dt = Double.valueOf(tfDT.getText()).doubleValue();
        sDT = tfDT.getText();
        applet.dt = dt;
        applet.dtInitial = dt;
      }
      catch (NumberFormatException ex) {
        JOptionPane.showMessageDialog(this,
            "Wrong input value format!\nPlease, try it again.",
            "Input error", JOptionPane.ERROR_MESSAGE);
        tfDT.setText(sDT);
        dt = Double.valueOf(sDT).doubleValue();
        allOK = false;
      }
      //we check if theta is in interval <0; 180> and r>0
      while (theta > 180)
        theta -= 360;
      while (theta < -180)
        theta += 360;
      if (theta < 0) {
        theta = -theta;
        phi = phi + 180;
        allOK = false;
      }
      while (phi > 360) {
        phi -= 360;
        allOK = false;
      }
      while (phi < 0) {
        phi += 360;
        allOK = false;
      }
      if (!allOK) {
        double phiPom = phi;
        if (phiPom > 180)
          phiPom = phi - 360;
        tfLongitude.setText(MyUtils.roundOff(phiPom, 4));
        tfLatitude.setText(MyUtils.roundOff(90 - theta, 4));
      }
      if (r - 6378e3 < 0) {
        r = 6378e3;
        tfH.setText(MyUtils.roundOff(0, 0));
        allOK = false;
      }

      if (allOK) {
        applet.setSimulation(r, theta, phi, vR, vTheta, vPhi, 0, dt);
        hide();
        applet.repaint();
      }
    }
    else {
      //after pressing Cancel button...
      tfH.setText(sH);
      tfLatitude.setText(sTheta);
      tfLongitude.setText(sPhi);
      tfVR.setText(sVR);
      tfVNorth.setText(sVTheta);
      tfVEast.setText(sVPhi);

      hide();
      applet.repaint();
    }
  }

}

class StepForwardHandler
    implements ActionListener {
  JOrbitLabOutline applet;

  public StepForwardHandler(JOrbitLabOutline applet) {
    this.applet = applet;
  }

  public void actionPerformed(ActionEvent e) {
    applet.simulationStep();
  }
}

class StepBackHandler
    implements ActionListener {
  JOrbitLabOutline applet;

  public StepBackHandler(JOrbitLabOutline applet) {
    this.applet = applet;
  }

  public void actionPerformed(ActionEvent e) {
    applet.odeAdaptiveSolver.setStepSize( -applet.dt);
    applet.simulationStep();
    applet.odeAdaptiveSolver.setStepSize(applet.dt);
  }
}

class Canvas
    extends JPanel {
  BufferedImage image;
  JOrbitLabOutline applet;

  //constructor
  public Canvas(BufferedImage image, JOrbitLabOutline applet) {
    this.image = image;
    this.applet = applet;
  }

  public void paintComponent(Graphics g) {
    Graphics2D g2 = (Graphics2D) g;
    g2.drawImage(image, null, 0, 0);
    applet.drawTopLayer(g2);
  }
}

class IntegerPoint {
  int i, j;

  public IntegerPoint(int i, int j) {
    this.i = i;
    this.j = j;
  }

  public int getI() {
    return i;
  }

  public int getJ() {
    return j;
  }
}

class Point3D {
  double x, y, z;

  public Point3D(double x, double y, double z) {
    this.x = x;
    this.y = y;
    this.z = z;
  }

  public double getX() {
    return x;
  }

  public double getY() {
    return y;
  }

  public double getZ() {
    return z;
  }
}

class ProjPoint {
  double phi, theta;

  public ProjPoint(double phi, double theta) {
    this.phi = phi;
    this.theta = theta;
  }

  public double getPhi() {
    return phi;
  }

  public double getTheta() {
    return theta;
  }

}

class Outline {
  int n; //current number of points in an outline
  IntegerPoint[] points;

  public Outline() {
    this.n = 0;
    this.points = new IntegerPoint[10000];
  }

  public void addPoint(IntegerPoint p) {
    points[n] = p;
    n++;
  }

  public IntegerPoint getPoint(int i) {
    if (i < n)
      return points[i];
    else
      return null;
  }
}